"""Support for bundles of related tasks."""

import secrets

from pydantic import BaseModel

from hipercow import ui
from hipercow.root import OptionalRoot, open_root
from hipercow.task import (
    TaskStatus,
    check_task_exists,
    check_task_id,
    task_status,
)


class Bundle(BaseModel):
    """A bundle of tasks.

    Attributes:
        name: The bundle name
        task_ids: The task identifiers in the bundle
    """

    name: str
    task_ids: list[str]


def bundle_create(
    task_ids: list[str],
    name: str | None = None,
    *,
    validate: bool = True,
    overwrite: bool = True,
    root: OptionalRoot = None,
) -> str:
    """Create a new bundle from a list of tasks.

    Arguments:
        task_ids: The task identifiers in the bundle

        name: The name for the bundle.  If not given, we randomly
            create one.  The format of the name is subject to change.

        validate: Check that all tasks exist before creating the bundle.

        overwrite: Overwrite a bundle if it already exists.

        root: The root, or if not given search from the current directory.

    Returns: The name of the newly created bundle.  Also, as a side
        effect, writes out the task bundle to disk.

    """
    root = open_root(root)
    if validate:
        for i in task_ids:
            check_task_exists(i, root)
    else:
        for i in task_ids:
            check_task_id(i)
    if name is None:
        # TODO: use something better here
        name = secrets.token_hex(8)

    path = root.path_bundle(name)
    if not overwrite and path.exists():
        msg = f"Bundle '{name}' exists and overwrite is False"
        raise Exception(msg)
    path.parent.mkdir(parents=True, exist_ok=True)
    bundle = Bundle(name=name, task_ids=task_ids)
    with path.open("w") as f:
        f.write(bundle.model_dump_json())
    ui.alert_success(f"Created bundle '{name}' with {len(task_ids)} tasks")
    return name


def bundle_load(name: str, root: OptionalRoot = None) -> Bundle:
    """Load a task bundle.

    Args:
        name: The name of the bundle to load
        root: The root, or if not given search from the current directory.

    Returns:
        The loaded bundle.
    """
    root = open_root(root)
    path = root.path_bundle(name)
    if not path.exists():
        msg = f"No such bundle '{name}'"
        raise Exception(msg)
    with path.open() as f:
        json_str = f.read()
    return Bundle.model_validate_json(json_str)


def bundle_list(root: OptionalRoot = None) -> list[str]:
    """List bundles.

    Args:
        root: The root, or if not given search from the current directory.

    Returns: The names of known bundles.  Currently the order of these
        is arbitrary.

    """
    root = open_root(root)
    path = root.path_bundle(None)
    # We could/should order by time here; we do a similar thing with
    # recent tasks.  Alternatively we might want a verbose mode that
    # stores this?  Or we could store time within the bundle, that
    # might make more sense but requiring serialising out as json,
    # possibly with pydantic?  Certainly this metadata would make
    # bundles more useful.
    nms = [x.name for x in path.glob("*")]
    return nms


def bundle_delete(name: str, root: OptionalRoot = None) -> None:
    """Delete a bundle.

    Note that this does not delete the tasks in the bundle, just the
    bundle itself.

    Args:
        name: The name of the bundle to delete
        root: The root, or if not given search from the current directory.

    Returns:
        Nothing, called for side effects only.

    """
    root = open_root(root)
    path = root.path_bundle(name)
    if not path.exists():
        msg = f"Can't delete bundle '{name}', it does not exist"
        raise Exception(msg)

    path.unlink()
    ui.alert_success("Deleted bundle '{name}'")


def bundle_status(name: str, root: OptionalRoot = None) -> list[TaskStatus]:
    """Get the statuses of tasks in a bundle.

    Depending on the context, `bundle_status_reduce()` may be more
    appropriate function to use, which attempts to reduce the list of
    statuses into the single "worst" status.

    Args:
        name: The name of the bundle to get the statuses for.
        root: The root, or if not given search from the current directory.

    Returns:
        A list of statuses, one per task.  These are stored in
        the same order as the original bundle.

    """
    bundle = bundle_load(name, root=root)
    return [task_status(i, root=root) for i in bundle.task_ids]


def bundle_status_reduce(name: str, root: OptionalRoot = None) -> TaskStatus:
    """Get the overall status from a bundle.

    Args:
        name: The name of the bundle to get the statuses for.
        root: The root, or if not given search from the current directory.

    Returns:
        The overall bundle status.
    """
    return _status_reduce(bundle_status(name, root))


def _status_reduce(status: list[TaskStatus]) -> TaskStatus:
    order = [
        TaskStatus.CREATED,
        TaskStatus.FAILURE,
        TaskStatus.CANCELLED,
        TaskStatus.RUNNING,
        TaskStatus.SUBMITTED,
        TaskStatus.SUCCESS,
    ]
    return order[min(order.index(i) for i in status)]


# Not implemented - result, cancel, wait, logs, retry



from pathlib import Path
from unittest import mock

import pytest
from taskwait import Result

import hipercow.dide.driver
from hipercow import root
from hipercow.dide import mounts
from hipercow.dide.configuration import DideConfiguration
from hipercow.dide.driver import ProvisionWaitWrapper, _dide_provision
from hipercow.dide.web import DideWebClient
from hipercow.resources import TaskResources
from hipercow.task import TaskStatus
from hipercow.util import transient_working_directory


def test_can_provision_with_dide(tmp_path, mocker):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with (r.path / "requirements.txt").open("w") as f:
        f.write("cowsay\n")

    m = mounts.Mount("host", "hostmount", Path("/local"))
    path_map = mounts.PathMap(tmp_path, m, "Q:", relative="path/to/dir")

    mock_client = mock.MagicMock(spec=DideWebClient)
    mocker.patch("hipercow.dide.driver._web_client", return_value=mock_client)
    mocker.patch("hipercow.dide.driver.taskwait")
    mocker.patch(
        "hipercow.dide.configuration.remap_path", return_value=path_map
    )
    config = DideConfiguration(
        r, mounts=[m], python_version=None, check_credentials=False
    )

    _dide_provision("myenv", "abcdef", config, r)
    resources = TaskResources(queue="BuildQueue")

    assert mock_client.submit.call_count == 1
    assert mock_client.mock_calls[0] == mock.call.submit(
        r"\\host\hostmount\path\to\dir\hipercow\py\env\myenv\provision\abcdef\run.bat",
        "myenv/abcdef",
        resources=resources,
    )

    assert hipercow.dide.driver.taskwait.call_count == 1
    assert hipercow.dide.driver.taskwait.mock_calls[0] == mock.call(mock.ANY)
    task = hipercow.dide.driver.taskwait.mock_calls[0].args[0]
    assert isinstance(task, ProvisionWaitWrapper)
    assert task.client == mock_client
    assert task.dide_id == mock_client.submit.return_value


def test_can_get_status_from_wait_wrapper():
    client = mock.MagicMock(spec=DideWebClient)
    client.status_job.side_effect = [TaskStatus.RUNNING, TaskStatus.SUCCESS]
    task = ProvisionWaitWrapper(mock.ANY, "myenv", "abcdef", client, "1234")
    assert task.status() == "running"
    assert task.status() == "success"
    assert client.status_job.call_count == 2


def test_wait_wrapper_can_get_log(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        name = "myenv"
        provision_id = "abcdef"
        task = ProvisionWaitWrapper(r, name, provision_id, mock.ANY, "1234")
        r.path_provision(name, provision_id).mkdir(parents=True)
        assert task.log() is None
        assert task.has_log()
        with r.path_provision_log(name, provision_id).open("w") as f:
            f.write("a\nb\n")
        assert task.log() == ["a", "b"]
        assert task.has_log()


def test_throw_after_failed_provision_with_dide(tmp_path, mocker, capsys):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with (r.path / "requirements.txt").open("w") as f:
        f.write("cowsay\n")

    m = mounts.Mount("host", "hostmount", Path("/local"))
    path_map = mounts.PathMap(tmp_path, m, "Q:", relative="path/to/dir")

    mock_client = mock.MagicMock(spec=DideWebClient)
    mock_client.log.return_value = "more logs"
    result = Result("failure", 100, 123)
    mocker.patch("hipercow.dide.driver._web_client", return_value=mock_client)
    mocker.patch("hipercow.dide.driver.taskwait", return_value=result)
    mocker.patch(
        "hipercow.dide.configuration.remap_path", return_value=path_map
    )
    config = DideConfiguration(
        r, mounts=[m], python_version=None, check_credentials=False
    )

    capsys.readouterr()
    with pytest.raises(Exception, match="Provisioning failed"):
        _dide_provision("myenv", "abcdef", config, r)
    out = capsys.readouterr().out

    assert "Provisioning failed after 23s!\n" in out
    assert "\nmore logs\n" in out

    assert mock_client.submit.call_count == 1
    assert mock_client.log.call_count == 1
    assert mock_client.log.mock_calls[0] == mock.call(
        mock_client.submit.return_value
    )
    assert hipercow.dide.driver.taskwait.call_count == 1



from pathlib import Path
from unittest import mock

import pytest

import hipercow.dide.check
from hipercow import root
from hipercow.dide.auth import Credentials
from hipercow.dide.check import (
    _dide_check_connection,
    _dide_check_credentials,
    _dide_check_path,
    _dide_check_root,
    _dide_check_root_configured,
    dide_check,
)
from hipercow.util import Result


def test_user_facing_check_function_success(mocker, capsys):
    ok = Result.ok()
    mocker.patch("hipercow.dide.check._dide_check_credentials", return_value=ok)
    mocker.patch("hipercow.dide.check._dide_check_connection", return_value=ok)
    mocker.patch("hipercow.dide.check._dide_check_path", return_value=ok)
    mocker.patch("hipercow.dide.check._dide_check_root", return_value=ok)
    dide_check(Path.cwd())
    out = capsys.readouterr().out
    assert "You look good to go!" in out


def test_user_facing_check_function_failure(mocker):
    ok = Result.ok()
    err = Result.err(Exception("foo"))
    mocker.patch("hipercow.dide.check._dide_check_credentials", return_value=ok)
    mocker.patch("hipercow.dide.check._dide_check_connection", return_value=ok)
    mocker.patch("hipercow.dide.check._dide_check_path", return_value=err)
    mocker.patch("hipercow.dide.check._dide_check_root", return_value=ok)
    with pytest.raises(Exception, match="You have issues to address"):
        dide_check()


def test_successful_credentials_read(mocker, capsys):
    creds = Credentials("bob", "password")
    mocker.patch("hipercow.dide.check.fetch_credentials", return_value=creds)
    mocker.patch("hipercow.dide.check.check_access")
    res = _dide_check_credentials()
    assert res

    out = capsys.readouterr().out
    assert "Checking DIDE credentials" in out
    assert "Found DIDE credentials" in out
    assert "Your username is 'bob'" in out
    assert "DIDE credentials are correct" in out


def test_error_if_no_credentials_found(mocker, capsys):
    mocker.patch(
        "hipercow.dide.check.fetch_credentials",
        side_effect=Exception("no creds"),
    )
    mocker.patch("hipercow.dide.check.check_access")
    res = _dide_check_credentials()
    assert not res

    out = capsys.readouterr().out
    assert "no creds\n" in out

    assert hipercow.dide.check.fetch_credentials.call_count == 1
    assert hipercow.dide.check.check_access.call_count == 0


def test_error_if_credentials_fail(mocker, capsys):
    creds = Credentials("bob", "password")
    mocker.patch("hipercow.dide.check.fetch_credentials", return_value=creds)
    mocker.patch(
        "hipercow.dide.check.check_access", side_effect=Exception("no access")
    )
    res = _dide_check_credentials()
    assert not res

    out = capsys.readouterr().out
    assert "no access\n" in out

    assert hipercow.dide.check.fetch_credentials.call_count == 1
    assert hipercow.dide.check.check_access.call_count == 1


def test_successful_connection_test(mocker, capsys):
    mocker.patch("hipercow.dide.check.requests")
    assert _dide_check_connection()

    out = capsys.readouterr().out
    assert "Connection to private network is working" in out


def test_failed_connection_test(mocker, capsys):
    mocker.patch(
        "hipercow.dide.check.requests.head",
        side_effect=Exception("no internet"),
    )
    assert not _dide_check_connection()
    out = capsys.readouterr().out
    assert "Failed to make connection to the private network" in out


def test_successful_path_mapping(mocker, capsys):
    mocker.patch("hipercow.dide.check.detect_mounts")
    mocker.patch("hipercow.dide.check.remap_path")
    assert _dide_check_path(Path.cwd())

    out = capsys.readouterr().out
    assert "Path looks like it is on a network share" in out


def test_failed_path_mapping(mocker, capsys):
    mocker.patch("hipercow.dide.check.detect_mounts")
    mocker.patch(
        "hipercow.dide.check.remap_path", side_effect=Exception("wrong path")
    )
    assert not _dide_check_path(Path.cwd())

    out = capsys.readouterr().out
    assert "Failed to map path to a network share" in out


def test_successful_initialisation(tmp_path, capsys, mocker):
    mock_check_configured = mocker.Mock()
    mocker.patch(
        "hipercow.dide.check._dide_check_root_configured", mock_check_configured
    )
    root.init(tmp_path)
    capsys.readouterr()
    assert _dide_check_root(tmp_path)
    out = capsys.readouterr().out
    assert "hipercow is initialised at" in out
    assert mock_check_configured.call_count == 1


def test_unsuccessful_initialisation(tmp_path, capsys):
    assert not _dide_check_root(tmp_path)
    out = capsys.readouterr().out
    assert "hipercow is not initialised" in out


def test_can_test_that_dide_windows_is_configured(capsys, mocker):
    mock_load_driver = mocker.Mock()
    mocker.patch("hipercow.dide.check.load_driver", mock_load_driver)
    root = mock.Mock()
    assert _dide_check_root_configured(root)
    out = capsys.readouterr().out
    assert "hipercow is configured to use 'dide-windows'" in out
    assert mock_load_driver.call_count == 1
    assert mock_load_driver.mock_calls[0] == mock.call("dide-windows", root)


def test_can_report_that_dide_windows_is_not_configured(capsys, mocker):
    mocker.patch(
        "hipercow.dide.check.load_driver",
        side_effect=Exception("not configured"),
    )
    root = mock.Mock()
    assert not _dide_check_root_configured(root)
    out = capsys.readouterr().out
    assert "hipercow is not configured to use 'dide-windows'" in out



import os
import platform
from pathlib import Path
from unittest import mock

import pytest

from hipercow.util import (
    check_python_version,
    expand_grid,
    find_file_descend,
    loop_while,
    subprocess_run,
    transient_envvars,
    transient_working_directory,
    truthy_envvar,
)


def test_find_descend(tmp_path):
    (tmp_path / "a" / "b" / "c" / "d").mkdir(parents=True)
    (tmp_path / "a" / "b" / ".foo").mkdir(parents=True)
    assert (
        find_file_descend(".foo", tmp_path / "a/b/c/d")
        == (tmp_path / "a" / "b").resolve()
    )
    assert find_file_descend(".foo", tmp_path / "a") is None


def test_transient_working_directory(tmp_path):
    here = Path.cwd()
    with transient_working_directory(None):
        assert Path.cwd() == here
    with transient_working_directory(tmp_path):
        assert Path.cwd() == tmp_path


def test_run_process_and_capture_output(tmp_path):
    path = tmp_path / "output"
    res = subprocess_run(["echo", "hello"], filename=path)
    assert res.returncode == 0
    assert path.exists()
    with open(path) as f:
        assert f.read().strip() == "hello"


def test_can_cope_with_missing_path(tmp_path, capsys):
    cmd = ["nosuchexe", "arg"]

    res = subprocess_run(cmd)
    assert res.args == cmd
    assert res.returncode == -1
    out = capsys.readouterr()
    assert len(out.out) > 0

    tmp = tmp_path / "log"
    res = subprocess_run(cmd, filename=tmp)
    assert capsys.readouterr().out == ""
    assert res.args == cmd
    assert res.returncode == -1

    with pytest.raises(FileNotFoundError):
        res = subprocess_run(cmd, check=True)


def test_can_set_envvars():
    with transient_envvars({"hc_a": "1"}):
        assert os.environ.get("hc_a") == "1"
    assert os.environ.get("hc_a") is None


def test_can_unset_envvars_if_none():
    with transient_envvars({"hc_a": "1", "hc_b": None, "hc_c": "3"}):
        assert os.environ.get("hc_a") == "1"
        assert "hc_b" not in os.environ
        assert os.environ.get("hc_c") == "3"
        with transient_envvars({"hc_b": "2", "hc_c": None}):
            assert os.environ.get("hc_a") == "1"
            assert os.environ.get("hc_b") == "2"
            assert os.environ.get("hc_c") is None
            assert "hc_c" not in os.environ
        assert os.environ.get("hc_a") == "1"
        assert "hc_b" not in os.environ
        assert os.environ.get("hc_c") == "3"


def test_can_check_python_version():
    ours3 = platform.python_version()
    ours2 = ".".join(ours3.split(".")[:2])
    assert check_python_version(None) == ours2
    assert check_python_version(ours3) == ours2
    assert check_python_version(ours2) == ours2
    assert check_python_version("3.10") == "3.10"
    assert check_python_version("3.11") == "3.11"
    with pytest.raises(Exception, match="does not parse as a valid version"):
        check_python_version("3.11a")
    with pytest.raises(Exception, match="is not supported"):
        check_python_version("3.9")
    assert check_python_version("3.9", ["3.9", "3.10"]) == "3.9"


def test_can_convert_envvar_to_bool():
    with transient_envvars({"hc_a": "1", "hc_b": "true", "hc_c": "TRUE"}):
        assert truthy_envvar("hc_a")
        assert truthy_envvar("hc_b")
        assert truthy_envvar("hc_c")
    with transient_envvars({"hc_a": None, "hc_b": "0", "hc_c": "t"}):
        assert not truthy_envvar("hc_a")
        assert not truthy_envvar("hc_b")
        assert not truthy_envvar("hc_c")


def test_loop_while_loops():
    fn = mock.Mock(side_effect=[True, True, False, False])
    loop_while(fn)
    assert fn.call_count == 3

    fn = mock.Mock(side_effect=[False, True, True, False, False])
    loop_while(fn)
    assert fn.call_count == 1


def test_expand_grid():
    assert expand_grid({}) == [{}]
    assert expand_grid({"a": [1]}) == [{"a": 1}]
    assert expand_grid({"a": [1, 2]}) == [{"a": 1}, {"a": 2}]
    assert expand_grid({"a": [1, 2], "b": [3]}) == [
        {"a": 1, "b": 3},
        {"a": 2, "b": 3},
    ]
    assert expand_grid({"a": [1, 2], "b": [3, 4, 5]}) == [
        {"a": 1, "b": 3},
        {"a": 1, "b": 4},
        {"a": 1, "b": 5},
        {"a": 2, "b": 3},
        {"a": 2, "b": 4},
        {"a": 2, "b": 5},
    ]



import platform
import re
import time
from pathlib import Path
from unittest import mock

import click
import pytest
from click.testing import CliRunner

from hipercow import cli, root, task
from hipercow.bundle import bundle_load
from hipercow.driver import list_drivers
from hipercow.resources import TaskResources
from hipercow.task import TaskData, TaskStatus, set_task_status
from hipercow.task_create import task_create_shell
from hipercow.util import transient_envvars
from tests.helpers import AnyInstanceOf


def test_can_init_repository(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        res = runner.invoke(cli.init, ".")
        assert res.exit_code == 0
        r = root.open_root()
        assert r.path == Path.cwd()
        res.stdout.startswith("Initialised hipercow at .")


def test_can_create_task(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        r = root.open_root()
        res = runner.invoke(cli.cli_task_create, ["echo", "hello", "world"])
        assert res.exit_code == 0
        task_id = res.stdout.strip()
        assert task.task_status(task_id, r) == task.TaskStatus.CREATED


def test_can_run_task(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        r = root.open_root()
        res = runner.invoke(cli.cli_task_create, ["echo", "hello", "world"])
        assert res.exit_code == 0
        task_id = res.stdout.strip()

        res = runner.invoke(cli.cli_task_status, task_id)
        assert res.exit_code == 0
        assert res.output.strip() == "created"

        res = runner.invoke(cli.cli_task_eval, task_id)
        assert res.exit_code == 0
        # It would be good to test that we get the expected output
        # here, and empirically we do.  However we don't seem to get
        # it captured in the runner output, though it is swallowed by
        # something.  I've checked with the capsys fixture and that
        # does not seem to have it either.
        assert task.task_status(task_id, r) == task.TaskStatus.SUCCESS


def test_can_save_and_read_log(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        res = runner.invoke(cli.cli_task_create, ["echo", "hello", "world"])
        task_id = res.stdout.strip()

        res = runner.invoke(cli.cli_task_log, task_id)
        assert res.exit_code == 0
        assert res.output == ""

        res = runner.invoke(cli.cli_task_eval, [task_id, "--capture"])
        assert res.exit_code == 0

        res = runner.invoke(cli.cli_task_log, task_id)
        assert res.exit_code == 0
        assert res.output == "hello world\n\n"

        res = runner.invoke(cli.cli_task_log, [task_id, "--outer"])
        assert res.exit_code == 1
        assert "outer logs are only available" in str(res.exception)


def test_can_process_with_status_args():
    assert cli._process_with_status([]) is None
    assert cli._process_with_status(["success"]) == TaskStatus.SUCCESS
    assert (
        cli._process_with_status(["success", "running"])
        == TaskStatus.RUNNING | TaskStatus.SUCCESS
    )


def test_can_list_tasks(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        res = runner.invoke(cli.cli_task_create, ["echo", "hello", "world"])
        assert res.exit_code == 0
        task_id = res.stdout.strip()

        res = runner.invoke(cli.cli_task_list, [])
        assert res.exit_code == 0
        assert res.output.strip() == task_id

        res = runner.invoke(cli.cli_task_list, ["--with-status", "created"])
        assert res.exit_code == 0
        assert res.output.strip() == task_id

        res = runner.invoke(cli.cli_task_list, ["--with-status", "running"])
        assert res.exit_code == 0
        assert res.output.strip() == ""


def test_can_call_cli_dide_authenticate(mocker):
    mocker.patch("hipercow.cli.dide_auth.check")
    mocker.patch("hipercow.cli.dide_auth.clear")
    mocker.patch("hipercow.cli.dide_auth.authenticate")

    runner = CliRunner()

    res = runner.invoke(cli.cli_dide_authenticate, [])
    assert res.exit_code == 0
    assert res.output.strip() == ""
    assert cli.dide_auth.check.call_count == 0
    assert cli.dide_auth.clear.call_count == 0
    assert cli.dide_auth.authenticate.call_count == 1

    res = runner.invoke(cli.cli_dide_authenticate, ["clear"])
    assert res.exit_code == 0
    assert res.output.strip() == ""
    assert cli.dide_auth.check.call_count == 0
    assert cli.dide_auth.clear.call_count == 1
    assert cli.dide_auth.authenticate.call_count == 1

    res = runner.invoke(cli.cli_dide_authenticate, ["check"])
    assert res.exit_code == 0
    assert res.output.strip() == ""
    assert cli.dide_auth.check.call_count == 1
    assert cli.dide_auth.clear.call_count == 1
    assert cli.dide_auth.authenticate.call_count == 1

    res = runner.invoke(cli.cli_dide_authenticate, ["other"])
    assert res.exit_code == 1
    assert "No such action 'other'" in str(res.exception)
    assert cli.dide_auth.check.call_count == 1
    assert cli.dide_auth.clear.call_count == 1
    assert cli.dide_auth.authenticate.call_count == 1


def test_can_run_dide_check(tmp_path, mocker):
    mock_check = mock.Mock()
    mocker.patch("hipercow.cli.dide_check", mock_check)
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        res = runner.invoke(cli.cli_dide_check, [])
        assert res.exit_code == 0
        assert mock_check.call_count == 1
        assert mock_check.mock_calls[0] == mock.call()


def test_can_configure_driver(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        res = runner.invoke(cli.init, ".")
        res = runner.invoke(cli.cli_driver_configure, ["example"])
        assert res.exit_code == 0
        assert "Configured hipercow to use 'example'" in res.output

        res = runner.invoke(cli.cli_driver_list, [])
        assert res.exit_code == 0
        assert res.output == "example\n"

        res = runner.invoke(cli.cli_driver_show, [])
        assert res.exit_code == 0
        assert "Configuration for 'example'" in res.output

        res = runner.invoke(cli.cli_driver_unconfigure, ["example"])
        assert res.exit_code == 0
        assert "Removed configuration for 'example'" in res.output

        res = runner.invoke(cli.cli_driver_list, [])
        assert res.exit_code == 0
        assert res.output == "(none)\n"


def test_can_list_environments(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        runner.invoke(cli.init, ".")
        runner.invoke(cli.cli_driver_configure, ["example"])
        res = runner.invoke(cli.cli_environment_new, [])
        assert res.exit_code == 0
        assert "Creating environment 'default' using 'pip'" in res.output
        res = runner.invoke(cli.cli_environment_list, [])
        assert res.exit_code == 0
        assert res.output == "default\nempty\n"


def test_can_create_task_in_environment(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        r = root.open_root()

        runner.invoke(cli.init, ".")
        res = runner.invoke(cli.cli_environment_new, ["--name", "other"])
        assert res.exit_code == 0

        res = runner.invoke(
            cli.cli_task_create,
            ["echo", "hello", "world", "--environment", "other"],
        )
        assert res.exit_code == 0
        task_id = res.stdout.strip()
        data = TaskData.read(task_id, r)
        assert data.environment == "other"


def test_can_provision_environment(tmp_path, mocker):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        runner.invoke(cli.init, ".")
        runner.invoke(cli.cli_environment_new, [])

        mock_provision = mock.MagicMock()
        mocker.patch("hipercow.cli.provision", mock_provision)

        res = runner.invoke(cli.cli_environment_provision, [])
        assert res.exit_code == 0
        assert mock_provision.call_count == 1
        assert mock_provision.mock_calls[0] == mock.call(
            "default", [], root=mock.ANY
        )

        res = runner.invoke(
            cli.cli_environment_provision, ["--name=foo", "pip", "install", "."]
        )
        assert res.exit_code == 0
        assert mock_provision.call_count == 2
        assert mock_provision.mock_calls[1] == mock.call(
            "foo", ["pip", "install", "."], root=mock.ANY
        )


def test_can_delete_environment(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        runner.invoke(cli.init, ".")
        runner.invoke(cli.cli_driver_configure, ["example"])
        res = runner.invoke(cli.cli_environment_new, ["--name", "other"])
        assert res.exit_code == 0
        assert "Creating environment 'other' using 'pip'" in res.output

        res = runner.invoke(cli.cli_environment_list, [])
        assert res.exit_code == 0
        assert res.output == "empty\nother\n"

        res = runner.invoke(cli.cli_environment_delete, ["--name", "other"])
        assert res.exit_code == 0

        res = runner.invoke(cli.cli_environment_list, [])
        assert res.exit_code == 0
        assert res.output == "empty\n"


def test_can_wait_on_task(tmp_path, mocker):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        runner.invoke(cli.init, ".")
        res = runner.invoke(cli.cli_task_create, ["echo", "hello", "world"])
        assert res.exit_code == 0
        task_id = res.stdout.strip()

        mocker.patch("hipercow.cli.task_wait")

        res = runner.invoke(cli.cli_task_wait, [task_id])
        assert res.exit_code == 0
        assert cli.task_wait.call_count == 1
        assert cli.task_wait.mock_calls[0] == mock.call(
            task_id,
            poll=1,
            timeout=None,
            show_log=True,
            progress=True,
        )

        res = runner.invoke(
            cli.cli_task_wait,
            [task_id, "--poll=0.1", "--no-show-log", "--timeout", "200"],
        )
        assert res.exit_code == 0
        assert cli.task_wait.call_count == 2
        assert cli.task_wait.mock_calls[1] == mock.call(
            task_id,
            poll=0.1,
            timeout=200,
            show_log=False,
            progress=True,
        )


def test_can_build_environment(tmp_path, mocker):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        runner.invoke(cli.init, ".")

        mock_provision = mock.MagicMock()
        mocker.patch("hipercow.cli.provision_run", mock_provision)

        res = runner.invoke(
            cli.cli_environment_provision_run, ["example", "abcdef"]
        )
        assert res.exit_code == 0
        assert mock_provision.call_count == 1
        assert mock_provision.mock_calls[0] == mock.call(
            "example", "abcdef", mock.ANY
        )


def test_can_create_on_task_and_wait(tmp_path, mocker):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        runner.invoke(cli.init, ".")
        mocker.patch("hipercow.cli.task_wait")
        res = runner.invoke(
            cli.cli_task_create, ["--wait", "echo", "hello", "world"]
        )
        assert res.exit_code == 0
        task_id = res.stdout.strip()
        assert cli.task_wait.call_count == 1
        assert cli.task_wait.mock_calls[0] == mock.call(
            task_id,
        )


def test_can_get_last_task(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        r = root.open_root()

        # No tasks
        res = runner.invoke(cli.cli_task_last, [])
        assert res.exit_code == 0
        assert res.stdout == ""

        res = runner.invoke(cli.cli_task_recent, [])
        assert res.exit_code == 0
        assert res.stdout == ""

        ids = [task_create_shell(["true"], root=r) for _ in range(5)]

        res = runner.invoke(cli.cli_task_last, [])
        assert res.exit_code == 0
        assert res.stdout == f"{ids[-1]}\n"

        res = runner.invoke(cli.cli_task_recent, [])
        assert res.exit_code == 0
        assert res.stdout == "".join(i + "\n" for i in ids)

        res = runner.invoke(cli.cli_task_recent, ["--limit", 2])
        assert res.exit_code == 0
        assert res.stdout == "".join(i + "\n" for i in ids[-2:])


def test_can_rebuild_recent_list(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        r = root.open_root()

        ids = []
        for i in range(5):
            if i > 0:
                time.sleep(0.01)
            ids.append(task_create_shell(["echo", "hello world"], root=r))

        res = runner.invoke(cli.cli_task_recent, [])
        assert res.exit_code == 0
        assert res.stdout == "".join(i + "\n" for i in ids)

        with r.path_recent().open("w") as f:
            for i in ids[:2] + [ids[2] + ids[3], ids[4]]:
                f.write(f"{i}\n")

        res = runner.invoke(cli.cli_task_recent, [])
        assert res.exit_code == 1
        assert res.stdout == ""
        assert (
            str(res.exception) == "Recent data list is corrupt, please rebuild"
        )

        res = runner.invoke(cli.cli_task_recent, ["--rebuild"])
        assert res.exit_code == 0
        assert res.stdout == "".join(i + "\n" for i in ids)


def test_can_call_cli_dide_bootstrap(mocker):
    mocker.patch("hipercow.cli.dide_bootstrap")
    runner = CliRunner()

    res = runner.invoke(cli.cli_dide_bootstrap, [])
    assert res.exit_code == 0
    assert res.output.strip() == ""
    assert cli.dide_bootstrap.call_count == 1
    assert cli.dide_bootstrap.mock_calls[0] == mock.call(
        None, force=False, verbose=True, python_versions=[]
    )

    res = runner.invoke(
        cli.cli_dide_bootstrap, ["myfile", "--verbose", "--force"]
    )
    assert res.exit_code == 0
    assert res.output.strip() == ""
    assert cli.dide_bootstrap.call_count == 2
    assert cli.dide_bootstrap.mock_calls[1] == mock.call(
        "myfile", force=True, verbose=True, python_versions=[]
    )


def test_show_exception(mocker):
    e = Exception("some error")
    mock_sys_exit = mock.MagicMock()
    mocker.patch("sys.exit", mock_sys_exit)
    with transient_envvars({"HIPERCOW_RAW_ERROR": "1"}):
        with pytest.raises(Exception, match="some error"):
            cli._handle_error(e)
    assert mock_sys_exit.call_count == 0


def test_show_small_error_only(capsys, mocker):
    e = Exception("some error")
    mock_sys_exit = mock.MagicMock()
    mock_console = mock.MagicMock()
    mocker.patch("sys.exit", mock_sys_exit)
    mocker.patch("hipercow.cli.console", mock_console)
    with transient_envvars(
        {"HIPERCOW_RAW_ERROR": None, "HIPERCOW_TRACEBACK": None}
    ):
        cli._handle_error(e)
    assert mock_sys_exit.call_count == 1
    assert mock_sys_exit.mock_calls[0] == mock.call(1)
    out = capsys.readouterr().out.splitlines()
    assert len(out) == 2
    assert out[0] == "Error: some error"
    assert out[1] == "For more information, run with 'HIPERCOW_TRACEBACK=1'"


def test_show_nice_traceback(capsys, mocker):
    e = Exception("some error")
    mock_sys_exit = mock.MagicMock()
    mock_console = mock.MagicMock()
    mocker.patch("sys.exit", mock_sys_exit)
    mocker.patch("hipercow.cli.console", mock_console)
    with transient_envvars(
        {"HIPERCOW_RAW_ERROR": None, "HIPERCOW_TRACEBACK": "1"}
    ):
        cli._handle_error(e)
    assert mock_sys_exit.call_count == 1
    assert mock_sys_exit.mock_calls[0] == mock.call(1)
    out = capsys.readouterr().out.splitlines()
    assert len(out) == 1
    assert out[0] == "Error: some error"
    assert len(mock_console.method_calls) == 1
    assert mock_console.method_calls[0] == mock.call.print_exception(
        show_locals=True, suppress=[click]
    )


def test_cli_wrapper_passes_to_exception_handler_on_error(mocker):
    e = Exception("some error")
    mock_handler = mocker.Mock()
    mocker.patch("hipercow.cli.cli", side_effect=e)
    mocker.patch("hipercow.cli._handle_error", mock_handler)
    cli.cli_safe()
    assert mock_handler.call_count == 1
    assert mock_handler.mock_calls[0] == mock.call(e)


def test_can_set_python_version(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        res = runner.invoke(cli.init, ".")

        r = root.open_root()
        v_ok = ".".join(platform.python_version_tuple()[:2])
        v_err = "3.11" if v_ok == "3.10" else "3.10"

        res = runner.invoke(
            cli.cli_driver_configure, ["example", "--python-version", v_err]
        )
        assert res.exit_code == 1
        assert "not the same as the local version" in str(res.exception)
        assert list_drivers(r) == []

        res = runner.invoke(
            cli.cli_driver_configure, ["example", "--python-version", v_ok]
        )
        assert res.exit_code == 0
        assert list_drivers(r) == ["example"]


def test_can_launch_repl(tmp_path, mocker):
    runner = CliRunner()
    mock_repl = mock.MagicMock()
    mocker.patch("hipercow.cli.repl", mock_repl)
    with runner.isolated_filesystem(temp_dir=tmp_path):
        res = runner.invoke(cli.cli_repl, [])
        assert res.exit_code == 0
        assert mock_repl.call_count == 1
        assert mock_repl.mock_calls[0] == mock.call(
            AnyInstanceOf(click.Context),
            prompt_kwargs={"message": "hipercow> ", "history": None},
        )

        runner.invoke(cli.init, ".")
        res = runner.invoke(cli.cli_repl, [])
        assert res.exit_code == 0
        assert mock_repl.call_count == 2
        assert mock_repl.mock_calls[1] == mock.call(
            AnyInstanceOf(click.Context),
            prompt_kwargs={
                "message": "hipercow> ",
                "history": AnyInstanceOf(cli.FileHistory),
            },
        )


def test_can_handle_exception_in_repl(mocker, capsys):
    ctx = mock.Mock()
    args = mock.Mock()
    mock_repl = mock.MagicMock(side_effect=[Exception("some error"), None])
    mocker.patch("hipercow.cli.repl", mock_repl)

    assert cli._repl_call(ctx, args)
    out = capsys.readouterr().out
    assert "Error: some error" in out
    assert mock_repl.call_count == 1
    assert mock_repl.mock_calls[0] == mock.call(ctx, prompt_kwargs=args)

    assert not cli._repl_call(ctx, args)
    out = capsys.readouterr().out
    assert out == ""
    assert mock_repl.call_count == 2
    assert mock_repl.mock_calls[1] == mock.call(ctx, prompt_kwargs=args)


def test_can_control_queue_used_in_task_creation(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        r = root.open_root()
        res = runner.invoke(cli.cli_driver_configure, ["example"])

        res = runner.invoke(
            cli.cli_task_create,
            ["echo", "hello", "world", "--queue", "default"],
        )
        assert res.exit_code == 0
        task_id = res.stdout.splitlines()[1]
        data = TaskData.read(task_id, r)
        assert data.resources == TaskResources(queue="default")

        res = runner.invoke(
            cli.cli_task_create,
            ["echo", "hello", "world", "--queue", "other"],
        )
        assert res.exit_code == 1
        assert "Queue 'other' is not in valid queue list" in str(res.exception)


def test_can_use_bulk_create(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        r = root.open_root()
        res = runner.invoke(
            cli.cli_create_bulk,
            [
                "--name",
                "mybundle",
                "--data",
                "a=1..2",
                "--data",
                "b=x,y",
                "mycmd",
                "@a",
                "c/@{b}/d",
            ],
        )
        assert res.exit_code == 0
        assert re.search("Created bundle '.+' with 4 tasks", res.output)
        bundle = bundle_load("mybundle", root=r)
        assert len(bundle.task_ids) == 4

        res = runner.invoke(cli.cli_bundle_list, [])
        assert res.exit_code == 0
        assert res.output.strip() == "mybundle"

        res = runner.invoke(cli.cli_bundle_delete, ["mybundle"])
        assert res.exit_code == 0

        res = runner.invoke(cli.cli_bundle_list, [])
        assert res.exit_code == 0
        assert res.output.strip() == ""


def test_can_get_bundle_status(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root.init(".")
        r = root.open_root()
        res = runner.invoke(
            cli.cli_create_bulk,
            [
                "--name",
                "mybundle",
                "--data",
                "a=1..5",
                "echo",
                "@a",
            ],
        )
        assert res.exit_code == 0
        assert re.search("Created bundle '.+' with 5 tasks", res.output)
        bundle = bundle_load("mybundle", root=r)

        set_task_status(bundle.task_ids[3], TaskStatus.SUCCESS, None, r)

        res = runner.invoke(cli.cli_bundle_status, ["mybundle"])
        assert res.exit_code == 0
        status = ["created", "created", "created", "success", "created"]
        assert res.output == "".join(
            f"{id}: {status}\n"
            for id, status in zip(bundle.task_ids, status, strict=False)
        )

        res = runner.invoke(
            cli.cli_bundle_status, ["mybundle", "--summary", "group"]
        )
        assert res.exit_code == 0
        assert res.output == "created: 4\nsuccess: 1\n"

        res = runner.invoke(
            cli.cli_bundle_status, ["mybundle", "--summary", "single"]
        )
        assert res.exit_code == 0
        assert res.output == "created\n"


def test_can_preview_commands():
    runner = CliRunner()
    res = runner.invoke(
        cli.cli_create_bulk,
        [
            "--preview",
            "--data",
            "a=1..3",
            "echo",
            "@a",
        ],
    )
    assert res.exit_code == 0
    lines = res.output.splitlines()
    assert "I would create 3 commands:" in lines[0]
    assert lines[1] == "  1: echo 1"
    assert lines[2] == "  2: echo 2"
    assert lines[3] == "  3: echo 3"


def test_can_summarise_generated_commands():
    runner = CliRunner()
    res = runner.invoke(
        cli.cli_create_bulk,
        [
            "--preview",
            "--data",
            "a=1..10",
            "--data",
            "b=1..5",
            "echo",
            "@a/@{b}",
        ],
    )
    assert res.exit_code == 0
    lines = res.output.splitlines()
    assert "I would create 50 commands:" in lines[0]
    assert lines[1] == "  1: echo 1/1"
    assert lines[2] == "  2: echo 1/2"
    assert lines[3] == "  3: echo 1/3"
    assert lines[4] == "   : ... 44 commands omitted"
    assert lines[5] == "  48: echo 10/3"
    assert lines[6] == "  49: echo 10/4"
    assert lines[7] == "  50: echo 10/5"


def test_can_create_commands_from_csv(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        with open("data.csv", "w") as f:
            f.write("a,b\n1,1\n1,2\n2,2")
        res = runner.invoke(
            cli.cli_create_bulk,
            [
                "--preview",
                "--data=data.csv",
                "echo",
                "@{a}-@{b}",
            ],
        )

        assert res.exit_code == 0
        output = res.output.splitlines()
        assert len(output) == 4
        assert output[1:] == ["  1: echo 1-1", "  2: echo 1-2", "  3: echo 2-2"]


def test_require_at_least_one_data_argument():
    runner = CliRunner()
    res = runner.invoke(
        cli.cli_create_bulk,
        [
            "--preview",
            "echo",
            "hello",
        ],
    )

    assert res.exit_code == 1
    assert str(res.exception) == "Expected at least one '--data' argument"


def test_can_parse_data_argument():
    assert cli._cli_bulk_parse_data_argument("a=1") == ("a", ["1"])
    assert cli._cli_bulk_parse_data_argument("a=1,2") == ("a", ["1", "2"])
    assert cli._cli_bulk_parse_data_argument("a=1:4") == ("a", ["1", "2", "3"])
    assert cli._cli_bulk_parse_data_argument("a=1..4") == (
        "a",
        ["1", "2", "3", "4"],
    )

    with pytest.raises(Exception, match="Failed to parse"):
        cli._cli_bulk_parse_data_argument("a=b=c")

    with pytest.raises(Exception, match="Failed to parse"):
        cli._cli_bulk_parse_data_argument("a")



import getpass
import re

import keyring

from hipercow import ui
from hipercow.dide.web import Credentials, check_access

## We follow hipercow in storing the username *as* a password, because
## it's largely a computer-specific things.


def authenticate():
    ui.h1("Please enter your DIDE credentials")
    ui.text(
        """
We need to know your DIDE username and password in order to log you into
the cluster. This will be shared across all projects on this machine, with
the username and password stored securely in your system keychain. You will
have to run this command again on other computers."""
    )

    ui.text(
        """
Your DIDE password may differ from your Imperial password, and in some
cases your username may also differ. If in doubt, perhaps try logging in
at https://mrcdata.dide.ic.ac.uk/hpc" and use the combination that works
for you there."""
    )

    ui.text(
        """
If you are unsure, please see our documentation about passwords:
https://mrc-ide.github.io/hipercow-py/dide/#about-our-usernames-and-passwords\
"""
    )

    ui.blank_line()
    username = _get_username(_default_username())

    ui.alert_info(f"Using username '{username}'")

    ui.blank_line()
    password = _get_password()

    ui.text(
        """
I am going to to try and log in with your password now.
If this fails we can always try again."""
    )

    credentials = Credentials(username, password)
    check_access(credentials)

    ui.blank_line()
    ui.alert_success("Username and password are correct")
    ui.alert_info("I am saving these into your keyring now")
    ui.alert_info(
        "You can delete your credentials with "
        "'hipercow dide authenticate clear'"
    )

    keyring.set_password("hipercow/dide/username", "", username)
    keyring.set_password("hipercow/dide/password", username, password)


def fetch_credentials() -> Credentials:
    username = keyring.get_password("hipercow/dide/username", "") or ""
    password = keyring.get_password("hipercow/dide/password", username)
    if not username or not password:
        # The error we throw here should depend on the context; if
        # we're within click then we should point people at at
        # 'hipercow dide authenticate' but if we are being used
        # programmatically that might not be best?
        msg = (
            "Did not find your DIDE credentials, "
            "please run 'hipercow dide authenticate'"
        )
        raise Exception(msg)
    return Credentials(username, password)


def check() -> None:
    ui.alert_info("Fetching credentials")
    credentials = fetch_credentials()
    ui.alert_info("Testing credentials")
    check_access(credentials)
    ui.alert_info("Success!")


def clear():
    username = keyring.get_password("hipercow/dide/username", "")
    if username:
        ui.alert_warning(f"Deleting credentials for '{username}'")
        _delete_password_silently("hipercow/dide/username", "")
        _delete_password_silently("hipercow/dide/password", username)


def _delete_password_silently(key: str, username: str):
    try:
        keyring.delete_password(key, username)
    except keyring.errors.PasswordDeleteError:
        pass


def _default_username() -> str:
    return (
        keyring.get_password("hipercow/dide/username", "") or getpass.getuser()
    )


# For mocking to work
def _get_input(text):
    return input(text)  # pragma: no cover


def _get_username(default: str) -> str:
    # I did look at the rich.prompt.Prompt() way of doing this but it
    # does not seem to add much:
    #
    #   value = Prompt.ask("DIDE username", default=default)
    #
    # with a similar implementation for the password case.
    value = _get_input(f"DIDE username (default: {default}) > ")
    return _check_username(value or default)


def _check_username(value) -> str:
    value = re.sub("^DIDE\\\\", "", value.strip(), flags=re.IGNORECASE)
    if not value:
        msg = "Invalid empty username"
        raise Exception(msg)
    if "\n" in value:
        msg = "Unexpected newline in username. Did you paste something?"
        raise Exception(msg)
    for char in "# ":
        if char in value:
            msg = f"Unexpected '{char}' in username"
            raise Exception(msg)
    return value


def _get_password() -> str:
    msg = (
        "Please enter your DIDE password. "
        "You will not see characters while you type."
    )
    ui.text(msg)
    value = getpass.getpass()
    if not value:
        msg = "Invalid empty password"
        raise Exception(msg)
    return value



import pickle

from hipercow import ui
from hipercow.dide.driver import DideWindowsDriver
from hipercow.driver import HipercowDriver, load_driver
from hipercow.example import ExampleDriver
from hipercow.root import OptionalRoot, Root, open_root
from hipercow.util import transient_working_directory


# For now, we'll hard code our two drivers (dide and example).  Later
# we can explore something like using hooks, for example in the style
# of pytest:
# * https://docs.pytest.org/en/stable/how-to/writing_plugins.html#pip-installable-plugins
# * https://packaging.python.org/en/latest/specifications/entry-points/
def configure(name: str, *, root: OptionalRoot = None, **kwargs) -> None:
    """Configure a driver.

    Configures a `hipercow` root to use a driver.

    Args:
        name: The name of the driver.  This will be `dide` unless you
            are developing `hipercow` itself :)
        root: The root, or if not given search from the current directory.
        **kwargs (Any): Arguments passed to, and supported by, your driver.

    Returns:
        Nothing, called for side effects only.
    """
    root = open_root(root)
    driver = _get_driver(name)
    with transient_working_directory(root.path):
        config = driver(root, **kwargs)
    _write_configuration(config, root)


def unconfigure(name: str, root: OptionalRoot = None) -> None:
    """Unconfigure (remove) a driver.

    Args:
        name: The name of the driver.  This will be `dide` unless you
            are developing `hipercow` itself :)
        root: The root, or if not given search from the current directory.

    Returns:
        Nothing, called for side effects only.
    """
    root = open_root(root)
    path = root.path_configuration(name)
    if path.exists():
        path.unlink()
        ui.alert_success(f"Removed configuration for '{name}'")
    else:
        ui.alert_warning(
            f"Did not remove configuration for '{name}' as it was not enabled"
        )


def show_configuration(
    name: str | None = None, root: OptionalRoot = None
) -> None:
    """Show a driver configuration.

    Args:
        name: The name of the driver.  This will be `dide` unless you
            are developing `hipercow` itself :)
        root: The root, or if not given search from the current directory.

    Returns:
        Nothing, called for side effects only.
    """
    root = open_root(root)
    dr = load_driver(name, root)
    ui.h1(f"Configuration for '{dr.name}'")
    dr.show_configuration()


# ahead of some sort of global store of drivers:
_DRIVERS = {d.name: d for d in [ExampleDriver, DideWindowsDriver]}


def _get_driver(name: str) -> type[HipercowDriver]:
    try:
        return _DRIVERS[name]
    except KeyError:
        msg = f"No such driver '{name}'"
        raise Exception(msg) from None


def _write_configuration(driver: HipercowDriver, root: Root) -> None:
    name = driver.name
    path = root.path_configuration(name)
    exists = path.exists()
    path.parent.mkdir(exist_ok=True, parents=True)
    with path.open("wb") as f:
        pickle.dump(driver, f)
    if exists:
        ui.alert_success(f"Updated configuration for '{name}'")
    else:
        ui.alert_success(f"Configured hipercow to use '{name}'")



import os
from unittest import mock

import pytest

from hipercow import root
from hipercow.environment import environment_engine, environment_new
from hipercow.environment_engines import Pip
from hipercow.util import file_create, transient_working_directory


def test_pip_environment_does_not_exist_until_created(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    env = Pip(r, "default")
    assert not env.exists()
    env.path().mkdir(parents=True)
    assert env.exists()


def test_pip_environment_can_be_created(tmp_path, mocker):
    mock_run = mock.MagicMock()
    mocker.patch("subprocess.run", mock_run)

    root.init(tmp_path)
    file_create(tmp_path / "requirements.txt")
    r = root.open_root(tmp_path)
    environment_new("default", "pip", r)
    env = environment_engine("default", r)
    assert isinstance(env, Pip)

    venv_path = str(env.path())
    envvars = env._envvars()

    env.create()
    assert mock_run.call_count == 1
    assert mock_run.mock_calls[0] == mock.call(
        ["python", "-m", "venv", venv_path], check=True, env=os.environ
    )

    with transient_working_directory(tmp_path):
        cmd = env.check_args(None)
        assert cmd == ["pip", "install", "--verbose", "-r", "requirements.txt"]
        env.provision(cmd)
        assert mock_run.call_count == 2
        assert mock_run.mock_calls[1] == mock.call(
            cmd,
            env=os.environ | envvars,
            check=True,
        )


def test_pip_can_determine_sensible_default_cmd(tmp_path):
    root.init(tmp_path)
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    env = Pip(r, "default")

    with transient_working_directory(tmp_path):
        with pytest.raises(Exception, match="Can't determine install"):
            env._auto()
        file_create(tmp_path / "requirements.txt")
        assert env._auto() == [
            "pip",
            "install",
            "--verbose",
            "-r",
            "requirements.txt",
        ]
        file_create(tmp_path / "pyproject.toml")
        assert env._auto() == ["pip", "install", "--verbose", "."]

        assert env.check_args(None) == ["pip", "install", "--verbose", "."]
        assert env.check_args([]) == ["pip", "install", "--verbose", "."]
        assert env.check_args(["pip", "install", "x"]) == [
            "pip",
            "install",
            "x",
        ]
        with pytest.raises(Exception, match="Expected first element"):
            assert env.check_args(["pwd"])



import base64
import datetime
import math
import re
from dataclasses import dataclass
from subprocess import list2cmdline
from urllib.parse import urljoin

import requests
from defusedxml import ElementTree

from hipercow.__about__ import __version__ as hipercow_version
from hipercow.resources import TaskResources
from hipercow.task import TaskStatus


def encode64(x: str) -> str:
    return base64.b64encode(x.encode("utf-8")).decode("utf-8")


def decode64(x: str) -> str:
    return base64.b64decode(x).decode("utf-8")


@dataclass
class Credentials:
    username: str
    password: str


@dataclass
class DideTaskStatus:
    dide_id: str
    name: str
    status: TaskStatus
    resources: str
    user: str
    time_start: float
    time_end: float
    time_submit: float
    template: str

    @staticmethod
    def from_string(entry):
        els = entry.strip().split("\t")
        els[2] = _parse_dide_status(els[2])
        els[4] = els[4].replace("DIDE\\", "")
        for i in range(5, 8):
            els[i] = _parse_dide_timestamp(els[i])
        return DideTaskStatus(*els)


class DideHTTPClient(requests.Session):
    _has_logged_in = False
    _credentials: Credentials

    def __init__(self, credentials: Credentials):
        super().__init__()
        self._credentials = credentials

    def request(self, method, path, *args, public=False, **kwargs):
        if not public and not self._has_logged_in:
            self.login()
        base_url = "https://mrcdata.dide.ic.ac.uk/hpc/"
        url = urljoin(base_url, path)
        headers = {"Accept": "text/plain"} if method == "POST" else {}
        response = super().request(
            method, url, *args, headers=headers, **kwargs
        )
        # To debug requests, you can do:
        # from requests_toolbelt.utils import dump
        # print(dump.dump_all(response).decode("utf-8"))
        response.raise_for_status()
        return response

    def login(self) -> None:
        data = {
            "us": encode64(self._credentials.username),
            "pw": encode64(self._credentials.password),
            "hpcfunc": encode64("login"),
        }
        res = self.request("POST", "index.php", data=data, public=True)
        no_access = "You don't seem to have any HPC access"
        if no_access in res.text:
            msg = "You do not have HPC access - please contact Wes"
            raise Exception(msg)
        self._has_logged_in = True

    def logout(self) -> None:
        self.request("GET", "logout.php", public=True)
        self._has_logged_in = False

    def username(self) -> str:
        return self._credentials.username

    def logged_in(self) -> bool:
        return self._has_logged_in


class DideWebClient:
    def __init__(self, credentials):
        self._client = DideHTTPClient(credentials)
        self._cluster = "wpia-hn"

    def login(self):
        self._client.login()

    def logout(self):
        self._client.logout()

    def headnodes(self) -> list[str]:
        data = {"user": encode64("")}
        response = self._client.request("POST", "_listheadnodes.php", data=data)
        return _client_parse_headnodes(response.text)

    def check_access(self) -> None:
        _client_check_access(self._cluster, self.headnodes())

    def logged_in(self) -> bool:
        return self._client.logged_in()

    def submit(
        self,
        path: str,
        name: str,
        resources: TaskResources,
        *,
        workdir: str | None = None,
    ) -> str:
        data = _client_body_submit(
            path, name, self._cluster, resources=resources, workdir=workdir
        )
        response = self._client.request("POST", "submit_1.php", data=data)
        return _client_parse_submit(response.text)

    def cancel(self, dide_id: str) -> bool:
        data = _client_body_cancel(dide_id, self._cluster)
        response = self._client.request("POST", "cancel.php", data=data)
        return _client_parse_cancel(response.text)

    def log(self, dide_id: str) -> str:
        data = _client_body_log(dide_id, self._cluster)
        response = self._client.request("POST", "showjobfail.php", data=data)
        return _client_parse_log(response.text)

    def status_user(self, state="*") -> list[DideTaskStatus]:
        data = _client_body_status_user(
            state, self._client.username(), self._cluster
        )
        response = self._client.request("POST", "_listalljobs.php", data=data)
        return _client_parse_status_user(response.text)

    def status_job(self, dide_id: str) -> TaskStatus:
        query = _client_query_status_job(dide_id, self._cluster)
        response = self._client.request("GET", "api/v1/get_job_status/", query)
        return _client_parse_status_job(response.text)

    def software(self):
        response = self._client.request(
            "GET", "api/v1/cluster_software", public=True
        )
        return _client_parse_software(response.json())


def check_access(credentials: Credentials) -> None:
    try:
        DideWebClient(credentials).check_access()
    except Exception as e:
        msg = "login failed"
        raise Exception(msg) from e


def _client_check_access(cluster: str, valid: list[str]) -> None:
    if cluster in valid:
        return
    if len(valid) == 0:
        msg = "You do not have access to any cluster"
    elif len(valid) == 1:
        msg = f"You do not have access to '{cluster}'; try '{valid[0]}'"
    else:
        valid_str = ", ".join(valid)
        msg = f"You do not have access to '{cluster}'; try one of {valid_str}"
    raise Exception(msg)


def _client_body_submit(
    path: str,
    name: str,
    cluster: str,
    *,
    resources: TaskResources,
    workdir: str | None,
) -> dict:
    # The str here keeps mypy happy, this will be a string by this
    # point.
    template = str(resources.queue)
    data = {
        "cluster": encode64(cluster),
        "template": encode64(template),
        "jn": encode64(name or ""),  # job name
        "wd": encode64(workdir or ""),  # work dir
        "se": encode64(""),  # stderr
        "so": encode64(""),  # stdout
        "jobs": encode64(_call_quote_batch_path(path)),
        "dep": encode64(""),  # dependencies, eventually
        "hpcfunc": "submit",
        "ver": encode64(f"hipercow-py/{hipercow_version}"),
    }

    if resources.cores == math.inf:
        data["rc"] = encode64("1")
        data["rt"] = encode64("Nodes")
    else:
        data["rc"] = encode64(str(resources.cores))
        data["rt"] = encode64("Cores")

    if resources.exclusive:
        data["exc"] = encode64("1")

    if resources.memory_per_task is not None:
        data["epm"] = encode64(str(1000 * resources.memory_per_task))

    if resources.memory_per_node is not None:
        data["mpn"] = encode64(str(1000 * resources.memory_per_node))

    if resources.max_runtime is not None:
        data["rnt"] = encode64(str(resources.max_runtime))

    # Still missing:
    #
    # hold until (hu)
    # requested nodes (rn)
    # priority (pri)

    return data


def _client_body_cancel(dide_id: str | list[str], cluster: str) -> dict:
    if isinstance(dide_id, str):
        dide_id = [dide_id]
    return {
        "cluster": encode64(cluster),
        "hpcfunc": encode64("cancel"),
        **{"c" + i: i for i in dide_id},
    }


def _client_body_log(dide_id: str, cluster: str) -> dict:
    return {"cluster": encode64(cluster), "hpcfunc": "showfail", "id": dide_id}


def _client_body_status_user(state: str, username: str, cluster: str) -> dict:
    return {
        "user": encode64(username),
        "scheduler": encode64(cluster),
        "state": encode64(state),
        "jobs": encode64("-1"),
    }


def _client_query_status_job(dide_id: str, cluster: str) -> dict:
    return {"scheduler": cluster, "jobid": dide_id}


def _client_parse_headnodes(txt: str) -> list[str]:
    txt = txt.strip()
    return txt.split("\n") if txt else []


def _client_parse_submit(txt: str) -> str:
    m = re.match("^Job has been submitted. ID: +([0-9]+)\\.$", txt.strip())
    if not m:
        msg = "Job submission has failed; could be a login error"
        raise Exception(msg)
    return m.group(1)


def _client_parse_cancel(txt: str):
    return dict([x.split("\t") for x in txt.strip().split("\n")])


def _client_parse_log(txt: str) -> str:
    res = ElementTree.fromstring(txt).find('.//input[@id="res"]')
    assert res is not None  # noqa: S101
    output = decode64(res.attrib["value"])
    return re.sub("^Output\\s*:\\s*?\n+", "", output)


def _client_parse_status_user(txt: str) -> list[DideTaskStatus]:
    return [DideTaskStatus.from_string(x) for x in txt.strip().split("\n")]


def _client_parse_status_job(txt: str) -> TaskStatus:
    return _parse_dide_status(txt.strip())


def _client_parse_software(json: dict) -> dict:
    # Likely to change soon when the portal updates
    linux = json["linuxsoftware"]
    windows = json["software"]

    def process(x):
        ret = {}
        for el in x:
            name = el["name"].lower()
            version = el["version"]
            omit = {"name", "version"}
            value = {k: v for k, v in el.items() if k not in omit}
            if name not in ret:
                ret[name] = {}
            ret[name][version] = value
        return ret

    return {"linux": process(linux), "windows": process(windows)}


def _parse_dide_status(status: str) -> TaskStatus:
    remap = {
        "Running": TaskStatus.RUNNING,
        "Finished": TaskStatus.SUCCESS,
        "Queued": TaskStatus.SUBMITTED,
        "Failed": TaskStatus.FAILURE,
        "Canceled": TaskStatus.CANCELLED,
        "Cancelled": TaskStatus.CANCELLED,
    }
    return remap[status]


def _parse_dide_timestamp(time: str) -> datetime.datetime:
    return datetime.datetime.strptime(time, "%Y%m%d%H%M%S").astimezone(
        datetime.timezone.utc
    )


def _call_quote_batch_path(path: str) -> str:
    # NOTE: list2cmdline is undocumented but needed.
    # not documented https://github.com/conan-io/conan/pull/11553/
    return f"call {list2cmdline([path])}"



from string import Template
from typing import TypeAlias

from hipercow.bundle import bundle_create
from hipercow.root import OptionalRoot, open_root
from hipercow.task_create import task_create_shell
from hipercow.util import expand_grid


class _TemplateAt(Template):
    delimiter = "@"

    # Backport a simplified version of get_idenfiers for python 3.10,
    # which lacks this method https://docs.python.org/3/library/string.html
    if not hasattr(Template, "get_identifiers"):  # pragma: no cover

        def get_identifiers(self):
            return _template_identifiers(self)


def _template_identifiers(obj: Template) -> list[str]:
    ids = []
    for mo in obj.pattern.finditer(obj.template):
        named = mo.group("named") or mo.group("braced")
        if named is not None and named not in ids:
            ids.append(named)
    return ids


BulkDataInput: TypeAlias = list[dict[str, str]] | dict[str, str | list[str]]


# ignoring the details of doing the substitutions, this is really all
# we need to do:
def bulk_create_shell(
    cmd_template: list[str],
    data: BulkDataInput,
    *,
    name: str | None = None,
    root: OptionalRoot = None,
    **kwargs,
) -> str:
    """Create a group of tasks from a template and data.

    You may want to use `bulk_crete_shell_commands` to preview the set
    of commands that would be created.

    There are a couple of ways that you might want to load data into
    the template:

    * You might have data in a table (e.g., in a csv file).  In this
      case we might model this as a list of `dict`s, each
      corresponding to a row in the original table.

    * You might have a series of factors that you want to compute all
      possible combinations of.  In this case we might model this as a
      dictionary of lists.

    We will handle either type and try and do the right thing.  This
    necessitates a bit of magic, and we may tone this down in future
    if it ends up being too unpredictable.

    Args:
        cmd_template: A command template.  This should be a list of
            strings, with some containing template placeholders.

        data: A list of dictionaries to substitute into the template.

        name: Optional name for the created bundle.  If `None` (the
            default) then a random name will be created for the bundle.

        root: The root, or if not given search from the current directory.

        kwargs: Additional arguments passed through to
            `hipercow.task_create.task_create_shell`.  This includes
            `environment`, `envvars`, `resources` and `driver`.

    Returns: The name of the created bundle of tasks.  You can use
        `hipercow.bundle.task_bundle_load` to load this and methods in
        `hipercow.bundle` to interact with or query the bundle.

    """
    root = open_root(root)
    cmd = bulk_create_shell_commands(cmd_template, data)
    task_ids = [task_create_shell(cmd_i, **kwargs) for cmd_i in cmd]
    return bundle_create(task_ids, name=name, validate=False, root=root)


def bulk_create_shell_commands(
    cmd_template: list[str], data: BulkDataInput
) -> list[list[str]]:
    """Create a list of commands from a template and data.

    Creates the list of commands (each of which is a list of strings)
    by using a **command template** -- a list of strings containing
    `@{...}` placeholders -- and data to substitute into this.

    You can use this to debug or preview commands that would be
    produced by `bulk_create_shell` without accidentally submitting
    thousands of tasks.

    Args:
        cmd_template: A command template.  This should be a list of
            strings, with some containing template placeholders.

        data: A list of dictionaries or dictionary of lists to
            substitute into the template; see
            `hipercow.task_create_bulk.bulk_create_shell` for details.

    Returns: A list of lists of strings; the `i`th element of this is
        the command substituted from the `i`th element of `data`.

    """
    template = [_TemplateAt(el) for el in cmd_template]

    # Check that all elements in data are consumed by the template

    keys_template = set()
    for el in template:
        keys_template |= set(el.get_identifiers())

    data_list = _check_template_data(data)
    keys_data = set(data_list[0].keys())

    if extra := keys_template - keys_data:
        extra_str = ", ".join(extra)
        msg = f"Template variables not present in data: {extra_str}"
        raise Exception(msg)

    if unused := keys_data - keys_template:
        unused_str = ", ".join(unused)
        msg = f"Data variables not present in template: {unused_str}"
        raise Exception(msg)

    return [[i.substitute(d) for i in template] for d in data_list]


def _check_template_data(data: BulkDataInput) -> list[dict[str, str]]:
    if not data:
        msg = "No data provided"
        raise Exception(msg)

    if isinstance(data, list):
        keys = data[0].keys()
        for el in data[1:]:
            if el.keys() != keys:
                msg = "Inconsistent keys among data"
                raise Exception(msg)
        return data
    else:
        return _bulk_data_combine(data)


def _bulk_data_combine(
    data: dict[str, str | list[str]],
) -> list[dict[str, str]]:
    return expand_grid(
        {k: v if isinstance(v, list) else [v] for k, v in data.items()}
    )



import pytest

from hipercow import root
from hipercow.configure import configure
from hipercow.environment import environment_new
from hipercow.provision import provision
from hipercow.task import TaskStatus, task_log, task_status
from hipercow.task_create import task_create_shell
from hipercow.task_eval import task_eval
from hipercow.util import transient_working_directory


@pytest.mark.slow
def test_run_in_environment(tmp_path):
    with transient_working_directory(tmp_path):
        root.init(tmp_path)
        r = root.open_root(tmp_path)

        with open("requirements.txt", "w") as f:
            f.write("cowsay\n")

        configure("example", root=r)
        environment_new("default", "pip", r)
        provision("default", [], root=r)
        tid = task_create_shell(["cowsay", "-t", "hello"], root=r)
        task_eval(tid, capture=True, root=r)

        assert task_status(tid, r) == TaskStatus.SUCCESS
        assert "| hello |" in task_log(tid, root=r)



from pathlib import Path
from subprocess import CompletedProcess
from unittest import mock

import pytest

from hipercow.dide import mounts


def test_can_parse_cifs_output():
    m = mounts._parse_unix_mount_entry(
        "//projects/other on /path/local type cifs (rw,relatime)"
    )
    assert m == mounts.Mount("projects", "other", Path("/path/local"))


def test_can_parse_mounts_on_unix(mocker):
    data = b"""//projects/other on /path/local type cifs (rw,relatime)
//projects/other2 on /path/local2 type cifs (rw,relatime)"""
    response = mock.MagicMock(spec=CompletedProcess)
    response.stdout = data
    mocker.patch("subprocess.run", return_value=response)
    res = mounts._detect_mounts_unix("Linux")
    assert len(res) == 2
    assert res[0] == mounts.Mount("projects", "other", Path("/path/local"))
    assert res[1] == mounts.Mount("projects", "other2", Path("/path/local2"))


def test_can_clean_dide_hostname():
    assert mounts._clean_dide_hostname("qdrive") == "qdrive"
    assert mounts._clean_dide_hostname("qdrive.dide.ic.ac.uk") == "qdrive"
    assert mounts._clean_dide_hostname("qdrive.other") == "qdrive.other"


def test_can_parse_mounts_on_windows(mocker):
    data = b"""#TYPE Microsoft.Management.Infrastructure.CimInstance#ROOT/Microsoft/Windows/SMB/MSFT_SmbMapping
"Status","LocalPath","RemotePath","RequireIntegrity","RequirePrivacy","UseWriteThrough","PSComputerName"
"OK","I:","\\\\wpia-hn\\hipercow","False","False","False",
"OK","Y:","\\\\wpia-hn2.hpc.dide.ic.ac.uk\\Climate","False","False","False",
"Disconnected","Z:","\\\\wpia-hn\\all-wpia-hn","False","False","False","""  # noqa: E501
    response = mock.MagicMock(spec=CompletedProcess)
    response.stdout = data
    mocker.patch("subprocess.run", return_value=response)
    res = mounts._detect_mounts_windows()
    assert len(res) == 2
    assert res[0] == mounts.Mount("wpia-hn", "hipercow", Path("I:/"))
    assert res[1] == mounts.Mount("wpia-hn2.hpc", "Climate", Path("Y:/"))


def test_can_get_correct_smb_type():
    assert mounts._unix_smb_mount_type("Linux") == "cifs"
    assert mounts._unix_smb_mount_type("Darwin") == "smbfs"


def test_throw_on_remap_with_no_mounts():
    with pytest.raises(Exception, match="Can't map local directory"):
        mounts.remap_path(Path.cwd(), [])


def test_can_remap_path():
    m = [mounts.Mount("host", "/hostmount", Path("/local"))]
    path = Path("/local/path/to/dir")
    res = mounts.remap_path(path, m)
    assert res == mounts.PathMap(path, m[0], "V:", relative="path/to/dir")


def test_throw_if_two_plausible_mounts():
    m = [
        mounts.Mount("host1", "/path1", Path("/local/path")),
        mounts.Mount("host2", "/path2", Path("/local")),
    ]
    path = Path("/local/path/to/dir")
    with pytest.raises(Exception, match="More than one plausible"):
        mounts.remap_path(path, m)


def test_preserve_drive_letter_if_given():
    m = [mounts.Mount("host", "/hostmount", Path("P:/"))]
    path = Path("P:/local/path")
    res = mounts.remap_path(path, m)
    assert res == mounts.PathMap(path, m[0], "P:", relative="local/path")


def test_can_map_home_to_q_drive():
    m = [mounts.Mount("qdrive", "user", Path("/local"))]
    path = Path("/local/path/to/dir")
    res = mounts.remap_path(path, m)
    assert res == mounts.PathMap(path, m[0], "Q:", relative="path/to/dir")


def test_can_parse_unix_entry():
    res = mounts._parse_unix_mount_entry(
        "//projects.dide.ic.ac.uk/other on /path/local type cifs (rw,relatime)"
    )
    assert res == mounts.Mount("projects", "other", Path("/path/local"))


def test_throw_if_error_in_unix_mount_entry():
    with pytest.raises(Exception, match="Failed to match mount output"):
        mounts._parse_unix_mount_entry("")


def test_can_parse_windows_mount_point():
    res = mounts._parse_windows_mount_entry(
        "E:", "//projects.dide.ic.ac.uk/other"
    )
    assert res == mounts.Mount("projects", "other", Path("E:/"))


def test_throw_if_error_in_windows_mount_point():
    with pytest.raises(Exception, match="Failed to parse windows entry"):
        mounts._parse_windows_mount_entry("", "E:")


def test_use_windows_detection_on_windows(mocker):
    mock_detect_windows = mock.Mock()
    mock_detect_unix = mock.Mock()
    mocker.patch("platform.system", return_value="Windows")
    mocker.patch(
        "hipercow.dide.mounts._detect_mounts_windows", mock_detect_windows
    )
    mocker.patch("hipercow.dide.mounts._detect_mounts_unix", mock_detect_unix)
    mounts.detect_mounts()
    assert mock_detect_windows.call_count == 1
    assert mock_detect_unix.call_count == 0


def test_use_windows_detection_on_unix(mocker):
    mock_detect_windows = mock.Mock()
    mock_detect_unix = mock.Mock()
    mocker.patch("platform.system", return_value="Linux")
    mocker.patch(
        "hipercow.dide.mounts._detect_mounts_windows", mock_detect_windows
    )
    mocker.patch("hipercow.dide.mounts._detect_mounts_unix", mock_detect_unix)
    mounts.detect_mounts()
    assert mock_detect_windows.call_count == 0
    assert mock_detect_unix.call_count == 1



from unittest import mock

from hipercow import root
from hipercow.configure import configure, show_configuration
from hipercow.dide.configuration import DideConfiguration
from hipercow.dide.mounts import Mount
from hipercow.dide.web import Credentials, DideWebClient
from hipercow.driver import list_drivers, load_driver
from hipercow.environment import environment_new
from hipercow.provision import provision
from hipercow.resources import TaskResources
from hipercow.task import task_log
from hipercow.task_create import task_create_shell
from hipercow.util import file_create, transient_working_directory


def test_can_configure_dide_mount(tmp_path, mocker):
    path = tmp_path / "a" / "b"
    root.init(path)
    r = root.open_root(path)
    mock_mounts = [Mount("projects", "other", tmp_path)]
    mock_check_auth = mock.Mock()
    mocker.patch("hipercow.dide.driver.detect_mounts", return_value=mock_mounts)
    mocker.patch("hipercow.dide.configuration.check_auth", mock_check_auth)
    configure("dide-windows", python_version=None, root=r)

    assert list_drivers(r) == ["dide-windows"]
    assert mock_check_auth.call_count == 1
    assert mock_check_auth.mock_calls[0] == mock.call()

    driver = load_driver("dide-windows", r)
    assert driver.config == DideConfiguration(
        r, mounts=mock_mounts, python_version=None, check_credentials=False
    )


def test_creating_task_triggers_submission(tmp_path, mocker):
    path = tmp_path / "a" / "b"
    root.init(path)
    r = root.open_root(path)
    mock_mounts = [Mount("projects", "other", tmp_path)]
    mock_creds = Credentials("bob", "secret")
    mock_web_client = mock.MagicMock(spec=DideWebClient)

    mocker.patch("hipercow.dide.driver.detect_mounts", return_value=mock_mounts)
    mocker.patch(
        "hipercow.dide.driver.fetch_credentials", return_value=mock_creds
    )
    mocker.patch("hipercow.dide.driver.DideWebClient", mock_web_client)
    mock_web_client.return_value.submit.return_value = "1234"

    configure(
        "dide-windows", python_version=None, root=r, check_credentials=False
    )
    with transient_working_directory(path):
        tid = task_create_shell(["echo", "hello world"], root=r)

    assert mock_web_client.call_count == 1
    assert mock_web_client.call_args == mock.call(mock_creds)
    cl = mock_web_client.return_value
    assert cl.login.call_count == 1
    assert cl.submit.call_count == 1
    # testing arguments here would be possibly useful, but we hit
    # issues with pathname normalisation very quickly.
    assert (r.path_task(tid) / "task_run.bat").exists()


def test_creating_task_with_resources(tmp_path, mocker):
    path = tmp_path / "a" / "b"
    root.init(path)
    r = root.open_root(path)
    mock_mounts = [Mount("projects", "other", tmp_path)]
    mock_creds = Credentials("bob", "secret")
    mock_web_client = mock.MagicMock(spec=DideWebClient)

    mocker.patch("hipercow.dide.driver.detect_mounts", return_value=mock_mounts)
    mocker.patch(
        "hipercow.dide.driver.fetch_credentials", return_value=mock_creds
    )
    mocker.patch("hipercow.dide.driver.DideWebClient", mock_web_client)
    mock_web_client.return_value.submit.return_value = "1234"

    configure(
        "dide-windows", python_version=None, check_credentials=False, root=r
    )
    resources = TaskResources(cores=4)
    with transient_working_directory(path):
        tid = task_create_shell(
            ["echo", "hello world"], resources=resources, root=r
        )

    assert mock_web_client.call_count == 1
    assert mock_web_client.call_args == mock.call(mock_creds)
    cl = mock_web_client.return_value
    assert cl.login.call_count == 1
    assert cl.submit.call_count == 1
    assert cl.submit.mock_calls[0][2]["resources"] == TaskResources(
        queue="AllNodes", cores=4
    )
    assert (r.path_task(tid) / "task_run.bat").exists()


def test_provision_using_driver(tmp_path, mocker):
    path = tmp_path / "a" / "b"
    root.init(path)
    r = root.open_root(path)
    mock_mounts = [Mount("projects", "other", tmp_path)]
    mock_provision = mock.MagicMock()
    mocker.patch("hipercow.dide.driver.detect_mounts", return_value=mock_mounts)
    mocker.patch("hipercow.dide.driver._dide_provision", mock_provision)
    configure(
        "dide-windows", python_version=None, check_credentials=False, root=r
    )
    environment_new("default", "pip", r)
    file_create(r.path / "requirements.txt")
    provision("default", [], root=r)
    cfg = load_driver(None, r).config
    assert mock_provision.call_count == 1
    assert mock_provision.mock_calls[0] == mock.call(
        "default", mock.ANY, cfg, r
    )


def test_resources_using_driver(tmp_path, mocker):
    path = tmp_path / "a" / "b"
    root.init(path)
    r = root.open_root(path)
    mock_mounts = [Mount("projects", "other", tmp_path)]
    mock_provision = mock.MagicMock()
    mocker.patch("hipercow.dide.driver.detect_mounts", return_value=mock_mounts)
    mocker.patch("hipercow.dide.driver._dide_provision", mock_provision)
    configure(
        "dide-windows", python_version=None, check_credentials=False, root=r
    )
    dr = load_driver(None, r)
    resources = dr.resources()
    assert resources.queues.default == "AllNodes"
    assert resources.max_cores == 32
    assert resources.max_memory == 512


def test_configure_python_version(tmp_path, mocker, capsys):
    path = tmp_path / "a" / "b"
    root.init(path)
    r = root.open_root(path)
    mock_mounts = [Mount("projects", "other", tmp_path)]
    mocker.patch("hipercow.dide.driver.detect_mounts", return_value=mock_mounts)
    configure(
        "dide-windows", python_version="3.12", check_credentials=False, root=r
    )
    capsys.readouterr()
    show_configuration(None, r)
    out = capsys.readouterr().out
    assert "Python version: 3.12" in out


def test_get_outer_logs_from_web_client(tmp_path, mocker):
    path = tmp_path / "a" / "b"
    root.init(path)
    r = root.open_root(path)
    mock_mounts = [Mount("projects", "other", tmp_path)]
    mock_creds = Credentials("bob", "secret")
    mock_web_client = mock.MagicMock(spec=DideWebClient)
    mocker.patch("hipercow.dide.driver.detect_mounts", return_value=mock_mounts)
    mocker.patch(
        "hipercow.dide.driver.fetch_credentials", return_value=mock_creds
    )
    mocker.patch(
        "hipercow.dide.driver.DideWebClient", return_value=mock_web_client
    )
    mock_web_client.submit.return_value = "1234"
    configure(
        "dide-windows", python_version=None, check_credentials=False, root=r
    )
    with transient_working_directory(path):
        tid = task_create_shell(["echo", "hello world"], root=r)

    assert task_log(tid, root=r) is None
    assert mock_web_client.log.call_count == 0

    res = task_log(tid, outer=True, root=r)
    assert res == mock_web_client.log.return_value
    assert mock_web_client.log.call_count == 1
    assert mock_web_client.log.mock_calls[0] == mock.call("1234")



import pytest

from hipercow import root
from hipercow.bundle import bundle_load
from hipercow.task import task_info
from hipercow.task_create_bulk import (
    _bulk_data_combine,
    _template_identifiers,
    _TemplateAt,
    bulk_create_shell,
    bulk_create_shell_commands,
)
from hipercow.util import transient_working_directory


def test_prepare_simple_grid():
    assert _bulk_data_combine({"a": ["0", "1", "2"]}) == [
        {"a": "0"},
        {"a": "1"},
        {"a": "2"},
    ]
    assert _bulk_data_combine({"a": ["0", "1", "2"], "b": "3"}) == [
        {"a": "0", "b": "3"},
        {"a": "1", "b": "3"},
        {"a": "2", "b": "3"},
    ]
    assert _bulk_data_combine({"a": ["0", "1", "2"], "b": ["3", "4"]}) == [
        {"a": "0", "b": "3"},
        {"a": "0", "b": "4"},
        {"a": "1", "b": "3"},
        {"a": "1", "b": "4"},
        {"a": "2", "b": "3"},
        {"a": "2", "b": "4"},
    ]


def test_can_construct_templated_calls():
    cmd = ["cmd", "path/@{a}", "@b"]
    pars = {"a": ["0", "1"], "b": ["2"]}
    data = _bulk_data_combine(pars)
    res = bulk_create_shell_commands(cmd, data)
    assert res == [["cmd", "path/0", "2"], ["cmd", "path/1", "2"]]
    assert bulk_create_shell_commands(cmd, pars) == res


def test_can_raise_if_unexpected_symbols_in_template():
    cmd = ["cmd", "path/@{a}", "@b"]
    pars = {"a": ["0", "1"]}
    pat = "Template variables not present in data: b"
    with pytest.raises(Exception, match=pat):
        bulk_create_shell_commands(cmd, pars)


def test_can_raise_if_unexpected_symbols_in_data():
    cmd = ["cmd", "path/@{a}"]
    pars = {"a": ["0", "1"], "b": ["2", "3"]}
    pat = "Data variables not present in template: b"
    with pytest.raises(Exception, match=pat):
        bulk_create_shell_commands(cmd, pars)


def test_can_raise_if_no_data():
    cmd = ["cmd", "path/@{a}"]
    with pytest.raises(Exception, match="No data provided"):
        bulk_create_shell_commands(cmd, {})
    with pytest.raises(Exception, match="No data provided"):
        bulk_create_shell_commands(cmd, [])


def test_can_raise_if_data_does_not_have_consistent_keys():
    cmd = ["cmd", "path/@{a}", "@b"]
    data = [{"a": "0", "b": "1"}, {"a": "0", "b": "1", "c": "2"}]
    with pytest.raises(Exception, match="Inconsistent keys among data"):
        bulk_create_shell_commands(cmd, data)


def test_can_bulk_create_tasks(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    data = {"by": ["cow", "horse"], "what": ["hello", "hipercow"]}
    with transient_working_directory(tmp_path):
        nm = bulk_create_shell(
            ["cowsay", "-c", "@by", "-t", "@{what}"],
            data,
        )
    bundle = bundle_load(nm, root=r)
    assert len(bundle.task_ids) == 4
    d0 = task_info(bundle.task_ids[0], root=r)
    assert d0.data.data["cmd"] == ["cowsay", "-c", "cow", "-t", "hello"]
    d1 = task_info(bundle.task_ids[1], root=r)
    assert d1.data.data["cmd"] == ["cowsay", "-c", "cow", "-t", "hipercow"]


def test_can_extract_identrifiers_with_backport():
    obj = _TemplateAt("hello @{a} @b world")
    assert _template_identifiers(obj) == ["a", "b"]
    obj = _TemplateAt("hello @{a} @b world @a")
    assert _template_identifiers(obj) == ["a", "b"]



import pickle
from abc import ABC, abstractmethod

from hipercow.resources import ClusterResources, TaskResources
from hipercow.root import Root
from hipercow.util import read_file_if_exists


class HipercowDriver(ABC):
    name: str

    @abstractmethod
    def __init__(self, root: Root, **kwargs):
        pass  # pragma: no cover

    @abstractmethod
    def show_configuration(self) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def submit(
        self, task_id: str, resources: TaskResources | None, root: Root
    ) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def provision(self, name: str, id: str, root: Root) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def resources(self) -> ClusterResources:
        pass  # pragma: no cover

    def task_log(
        self, task_id: str, *, outer: bool = False, root: Root
    ) -> str | None:
        if outer:
            return None
        return read_file_if_exists(root.path_task_log(task_id))


def list_drivers(root) -> list[str]:
    path = root.path_configuration(None)
    return [x.name for x in path.glob("*")]


def load_driver(driver: str | None, root: Root) -> HipercowDriver:
    dr = _load_driver(driver, root)
    if not dr:
        msg = "No driver configured"
        raise Exception(msg)
    return dr


def load_driver_optional(
    driver: str | None,
    root: Root,
) -> HipercowDriver | None:
    return _load_driver(driver, root)


def _load_driver(driver: str | None, root: Root) -> HipercowDriver | None:
    if not driver:
        return _default_driver(root)
    path = root.path_configuration(driver)
    if not path.exists():
        msg = f"No such driver '{driver}'"
        raise Exception(msg)
    with open(path, "rb") as f:
        return pickle.load(f)


def _default_driver(root: Root) -> HipercowDriver | None:
    candidates = list_drivers(root)
    n = len(candidates)
    if n == 0:
        return None
    if n > 1:
        msg = "More than one candidate driver"
        raise Exception(msg)
    return load_driver(candidates[0], root)



"""Some quality of life printing utilities, ported from R's cli."""

from rich.console import Console

console = Console()


def h1(text: str) -> None:
    before = "─" * 3
    after = "─" * (max(console.width - len(before) - 3 - len(text), 0))
    console.print(
        f"[cyan]{before}[/cyan] [bold]{text}[/bold] [cyan]{after}[/cyan]"
    )


def text(text: str, **kwargs) -> None:
    console.print(text, **kwargs)


def logs(title: str, text: str | None, *, indent: int = 0) -> None:
    indent_str = " " * indent
    alert_arrow(title, indent=indent)
    if text is None:
        console.print(f"{indent_str}(no logs found)", style="dim")
    else:
        console.rule(style="dim")
        console.print(text, emoji=False, markup=False, crop=False)
        console.rule(style="dim")


def li(text: str, indent: int = 0, symbol: str = "*", title: str = ""):
    indent_str = " " * indent
    title = f"[bold]{title}[/bold] " if title else ""
    console.print(f"{indent_str}{symbol} {title}{text}")


def blank_line(n: int = 1) -> None:
    console.print("\n" * (n - 1))


def alert_danger(text: str, indent: int = 0) -> None:
    alert(":heavy_multiplication_x:", text, "bold red", indent=indent)


def alert_success(text: str, indent: int = 0) -> None:
    alert(":heavy_check_mark:", text, "bold green", indent=indent)


def alert_warning(text: str, indent: int = 0) -> None:
    alert("!", text, "bold orange", indent=indent)


def alert_info(text: str, indent: int = 0) -> None:
    alert("i", text, "bold cyan", indent=indent)


def alert_see_also(
    text: str, prefix: str = "For more information, see ", indent: int = 0
) -> None:
    alert(":books:", f"{prefix}{text}", indent=indent)


def alert_arrow(text: str, indent: int = 0) -> None:
    alert(":arrow_forward:", text, "bold yellow", indent=indent)


def alert(
    icon: str, text: str, style: str | None = None, indent: int = 0
) -> None:
    indent_str = " " * indent
    if style:
        console.print(f"{indent_str}[{style}]{icon}[/{style}] {text}")
    else:
        console.print(f"{indent_str}{icon} {text}")



from pathlib import Path
from unittest import mock

import pytest

from hipercow.dide.bootstrap import (
    BootstrapTask,
    _bootstrap_args,
    _bootstrap_check_pipx_pyz,
    _bootstrap_mount,
    _bootstrap_python_versions,
    _bootstrap_submit,
    _bootstrap_target,
    _bootstrap_unc,
    _bootstrap_wait,
    bootstrap,
)
from hipercow.dide.mounts import Mount
from hipercow.dide.web import DideWebClient
from hipercow.resources import TaskResources
from hipercow.util import file_create


def test_can_construct_unc_paths():
    assert _bootstrap_unc(Path("a/b")) == r"\\wpia-hn\hipercow\a\b"


def test_can_build_args():
    assert _bootstrap_args(force=False, verbose=False) == ""
    assert _bootstrap_args(force=True, verbose=False) == "--force"
    assert _bootstrap_args(force=False, verbose=True) == "--verbose"
    assert _bootstrap_args(force=True, verbose=True) == "--force --verbose"


def test_can_detect_bootstrap_mount():
    mounts = [
        Mount(host="foo", remote="path", local=Path("a")),
        Mount(host="bar", remote="hipercow", local=Path("b")),
        Mount(host="wpia-hn.hpc", remote="hipercow", local=Path("c")),
    ]
    assert _bootstrap_mount(mounts) == mounts[2]
    with pytest.raises(Exception, match="Failed to find"):
        _bootstrap_mount(mounts[:1])


def test_can_use_pypi_hipercow():
    assert _bootstrap_target(None, None, "abcdef") == "hipercow"


def test_can_use_local_hipercow(tmp_path):
    src = tmp_path / "src"
    mount = Mount(host="wpia-hn.hpc", remote="hipercow", local=tmp_path / "dst")
    with pytest.raises(FileNotFoundError):
        _bootstrap_target(str(src), mount, "abcdef")
    with src.open("w") as f:
        f.write("contents\n")
    res = _bootstrap_target(str(src), mount, "abcdef")
    assert res == r"\\wpia-hn\hipercow\bootstrap-py-windows\in\abcdef\src"
    dst = tmp_path / "dst/bootstrap-py-windows/in/abcdef/src"
    assert dst.exists()
    with dst.open() as f:
        assert f.read() == "contents\n"


def test_can_submit_bootstrap_task(tmp_path):
    client = mock.MagicMock(spec=DideWebClient)
    mount = Mount(host="wpia-hn.hpc", remote="hipercow", local=tmp_path)
    bootstrap_id = "abcdef"
    version = "3.11"
    target = "hipercow"
    args = ""
    t = _bootstrap_submit(client, mount, bootstrap_id, version, target, args)
    resources = TaskResources(queue="AllNodes")
    assert t.client == client
    assert client.submit.call_count == 1
    assert client.submit.mock_calls[0] == mock.call(
        r"\\wpia-hn\hipercow\bootstrap-py-windows\in\abcdef\3.11.bat",
        "bootstrap/abcdef/3.11",
        resources,
    )
    assert t.dide_id == client.submit.return_value
    dest = tmp_path / "bootstrap-py-windows/in/abcdef/3.11.bat"
    assert dest.exists()
    with dest.open() as f:
        contents = f.readlines()
    assert "call set_python_311_64\n" in contents

    assert t.log() is None
    assert not t.has_log()

    status = t.status()
    assert client.status_job.call_count == 1
    assert client.status_job.mock_calls[0] == mock.call(t.dide_id)
    assert status == str(client.status_job.return_value)


def test_can_wait_on_successful_tasks(tmp_path, capsys):
    client = mock.MagicMock(spec=DideWebClient)
    client.status_job.return_value = "success"
    mount = Mount(host="wpia-hn.hpc", remote="hipercow", local=tmp_path)
    bootstrap_id = "abcdef"
    tasks = [
        BootstrapTask(mount, bootstrap_id, client, "1", "3.11"),
        BootstrapTask(mount, bootstrap_id, client, "2", "3.12"),
    ]
    _bootstrap_wait(tasks)
    out = capsys.readouterr().out
    assert "Waiting on 2 tasks" in out
    assert "3.11: success" in out
    assert "3.12: success" in out


def test_can_error_on_failed_tasks(tmp_path, capsys):
    client = mock.MagicMock(spec=DideWebClient)
    client.status_job.side_effect = ["success", "failure"]
    client.log.return_value = "some log"
    mount = Mount(host="wpia-hn.hpc", remote="hipercow", local=tmp_path)
    bootstrap_id = "abcdef"
    tasks = [
        BootstrapTask(mount, bootstrap_id, client, "1011", "3.11"),
        BootstrapTask(mount, bootstrap_id, client, "1012", "3.12"),
    ]
    tasks[1].path_log.parent.mkdir(parents=True)
    with tasks[1].path_log.open("w") as f:
        f.write("log1\nlog2\n")
    with pytest.raises(Exception, match="1/2 bootstrap tasks failed"):
        _bootstrap_wait(tasks)
    out = capsys.readouterr().out
    assert "3.11: success" in out
    assert "3.12: failure" in out
    assert "Additional logs from cluster for task '1012':" in out
    assert "\nsome log\n" in out
    assert "Logs from pipx:" in out
    assert "\nlog1\nlog2\n" in out


# This test is revolting:
def test_can_launch_bootstrap(mocker):
    mock_client = mock.MagicMock(spec=DideWebClient)
    mock_mount = mock.MagicMock()
    mock_submit = mock.MagicMock()
    mock_wait = mock.MagicMock()
    mock_rmtree = mock.MagicMock()

    mocker.patch("hipercow.dide.bootstrap._web_client", mock_client)
    mocker.patch("hipercow.dide.bootstrap._bootstrap_mount", mock_mount)
    mocker.patch("hipercow.dide.bootstrap._bootstrap_submit", mock_submit)
    mocker.patch("hipercow.dide.bootstrap._bootstrap_wait", mock_wait)
    mocker.patch("shutil.rmtree", mock_rmtree)
    bootstrap(None)

    assert mock_client.call_count == 1
    assert mock_mount.call_count == 1
    assert mock_submit.call_count == 4

    client = mock_client.return_value
    mount = mock_mount.return_value

    assert mock_submit.mock_calls[0] == mock.call(
        client, mount, mock.ANY, "3.10", "hipercow", ""
    )
    assert mock_wait.call_count == 1
    assert len(mock_wait.mock_calls[0].args[0]) == 4
    assert mock_wait.mock_calls[0].args[0][3] == mock_submit.return_value


def test_error_if_no_pipx_pyz(tmp_path):
    with pytest.raises(Exception, match="Expected 'pipx.pyz' to be found"):
        _bootstrap_check_pipx_pyz(tmp_path)
    file_create(tmp_path / "pipx.pyz")
    assert _bootstrap_check_pipx_pyz(tmp_path) is None


def test_can_set_versions_to_install():
    assert _bootstrap_python_versions(["3.13"]) == ["3.13"]
    assert _bootstrap_python_versions(None) == ["3.10", "3.11", "3.12", "3.13"]



from dataclasses import dataclass
from pathlib import Path

import requests

from hipercow import ui
from hipercow.dide.auth import Credentials, check_access, fetch_credentials
from hipercow.dide.mounts import detect_mounts, remap_path
from hipercow.driver import load_driver
from hipercow.root import Root, open_root
from hipercow.util import Result

_DOCS = "https://mrc-ide.github.io/hipercow-py"


@dataclass
class DideCheckResult:
    credentials: Result
    connection: Result
    path: Result
    root: Result

    def __bool__(self) -> bool:
        return (
            bool(self.credentials) and bool(self.connection) and bool(self.path)
        )


def dide_check(path: Path | None = None) -> None:
    """Check that everything is correctly configured to use hipercow."""
    path = path or Path.cwd()

    result = DideCheckResult(
        _dide_check_credentials(),
        _dide_check_connection(),
        _dide_check_path(path),
        _dide_check_root(path),
    )
    if not result:
        msg = "You have issues to address before using hipercow"
        raise Exception(msg)
    ui.alert_success("You look good to go!")


def _dide_check_credentials() -> Result:
    ui.alert_arrow("Checking DIDE credentials")
    try:
        credentials = fetch_credentials()
        ui.alert_success("Found DIDE credentials", indent=4)
        ui.alert_info(f"Your username is '{credentials.username}'", indent=4)
        return _dide_check_access(credentials)
    except Exception as e:
        ui.alert_danger(str(e), indent=4)
        ui.alert_see_also(f"{_DOCS}/dide/#authentication-with-dide", indent=4)
        return Result.err(e)


def _dide_check_access(credentials: Credentials) -> Result:
    try:
        # TODO: we should break this up a little more and separate out
        # login-failure from access-failure.
        check_access(credentials)
        ui.alert_success("DIDE credentials are correct", indent=4)
        return Result.ok()
    except Exception as e:
        ui.alert_danger(str(e), indent=4)
        ui.alert_see_also(f"{_DOCS}/dide/#authentication-with-dide", indent=4)
        return Result.err(e)


def _dide_check_connection() -> Result:
    ui.alert_arrow("Checking network connections")
    try:
        url = "https://vault.dide.ic.ac.uk:8200"
        requests.head(url, timeout=1)
        ui.alert_success("Connection to private network is working", indent=4)
        return Result.ok()
    except Exception as e:
        ui.alert_danger(
            "Failed to make connection to the private network", indent=4
        )
        ui.alert_info("Please check that you have ZScalar enabled", indent=4)
        ui.alert_see_also(f"{_DOCS}/dide/#networks", indent=4)
        return Result.err(e)


def _dide_check_path(path: Path) -> Result:
    ui.alert_arrow("Checking paths")
    mounts = detect_mounts()
    try:
        map = remap_path(path, mounts)
        ui.alert_success("Path looks like it is on a network share", indent=4)
        ui.alert_info(f"Using '{path}'", indent=4)
        ui.alert_info(
            f"This is '{map.mount.remote}' on '{map.mount.host}'", indent=4
        )
        return Result.ok()
    except Exception as e:
        ui.alert_danger("Failed to map path to a network share", indent=4)
        ui.alert_info(str(e), indent=4)
        ui.alert_see_also(f"{_DOCS}/dide/#filesystems-and-paths", indent=4)
        return Result.err(e)


def _dide_check_root(path: Path) -> Result:
    ui.alert_arrow("Checking hipercow root")
    try:
        root = open_root(path)
        ui.alert_success(f"hipercow is initialised at '{root.path}'", indent=4)
        return _dide_check_root_configured(root)
    except Exception as e:
        ui.alert_danger("hipercow is not initialised", indent=4)
        ui.alert_info(
            "You can run 'hipercow init' to initialise the root", indent=4
        )
        ui.alert_see_also(f"{_DOCS}/introduction/#initialisation", indent=4)
        return Result.err(e)


def _dide_check_root_configured(root: Root) -> Result:
    try:
        load_driver("dide-windows", root)
        ui.alert_success(
            "hipercow is configured to use 'dide-windows'", indent=4
        )
        return Result.ok()
    except Exception as e:
        ui.alert_danger(
            "hipercow is not configured to use 'dide-windows'", indent=4
        )
        ui.alert_info(
            "You can run 'hipercow driver configure dide-windows' to configure the root",  # noqa: E501
            indent=4,
        )
        ui.alert_see_also(f"{_DOCS}/introduction/#initialisation", indent=4)
        return Result.err(e)



class AnyInstanceOf:
    def __init__(self, cls):
        self.cls = cls

    def __eq__(self, other):
        return isinstance(other, self.cls)

    def __ne__(self, other):
        return not isinstance(other, self.cls)

    def __repr__(self):
        return f"<ANY {self.cls.__name__}>"



import os
from unittest import mock

import pytest

from hipercow import root
from hipercow.environment import environment_engine
from hipercow.environment_engines import Empty


def test_empty_environment_has_no_path_but_always_exists(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    env = Empty(r, "empty")

    assert env.exists()
    with pytest.raises(Exception, match="The empty environment has no path"):
        env.path()


def test_can_load_empty_engine(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    assert isinstance(environment_engine("empty", r), Empty)
    assert isinstance(environment_engine("default", r), Empty)


def test_empty_environment_cannot_be_created(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    env = Empty(r, "empty")
    with pytest.raises(Exception, match="Can't create the empty environment!"):
        env.create()


def test_empty_environment_cannot_be_provisioned(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    env = Empty(r, "empty")
    with pytest.raises(Exception, match="Can't provision the empty"):
        env.provision([])


def test_can_run_command_in_empty_env(tmp_path, mocker):
    mock_run = mock.MagicMock()
    mocker.patch("subprocess.run", mock_run)
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    env = Empty(r, "empty")
    env.run(["some", "command"])
    assert mock_run.call_count == 1
    assert mock_run.mock_calls[0] == mock.call(
        ["some", "command"], env=os.environ, check=False
    )


def test_can_validate_empty_args(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    env = environment_engine("empty", r)
    assert env.check_args(None) == []
    assert env.check_args([]) == []
    with pytest.raises(Exception, match="No arguments are allowed"):
        env.check_args(["other"])



"""Create Python virtual environments using pip."""

import os
import subprocess
from pathlib import Path

from hipercow.environment_engines.base import EnvironmentEngine, Platform
from hipercow.root import Root
from hipercow.util import subprocess_run, transient_envvars


class Pip(EnvironmentEngine):
    """Python virtual environments, installed by pip."""

    def __init__(self, root: Root, name: str, platform: Platform | None = None):
        super().__init__(root, name, platform)

    def path(self) -> Path:
        return super().path() / f"venv-{self.platform.system}"

    def create(self, **kwargs) -> None:
        """Create the virtual environment.

        Calls

        ```
        python -m venv <path>
        ```

        with the result of `path()`.

        Args:
            **kwargs (Any): Additional arguments to `subprocess_run`

        Returns:
            Nothing, called for side effects only.
        """
        cmd = ["python", "-m", "venv", str(self.path())]
        subprocess_run(cmd, check=True, **kwargs)

    def check_args(self, cmd: list[str] | None) -> list[str]:
        """Validate pip installation command.

        Checks if `cmd` is a valid `pip` command.

        If `cmd` is `None` or the empty list, we try and guess a
        default command, based on files found in your project root.

        * if you have a `pyproject.toml` file, then we will try and
          run `pip install --verbose .`

        * if you have a `requirements.txt`, then we will try and run
          `pip install --verbose -r requirements.txt`

        (In both cases these are returned as a list of arguments.)

        If there are other reasonable conventions that we might
        follow, please let us know.

        Args:
            cmd: The command to validate

        Returns:
            A validated list of arguments.
        """
        if not cmd:
            return self._auto()
        if cmd[0] != "pip":
            msg = "Expected first element of 'cmd' to be 'pip'"
            raise Exception(msg)
        return cmd

    def provision(self, cmd: list[str], **kwargs) -> None:
        """Provision a virtual environment using pip.

        Args:
            cmd: The command to run

            **kwargs (Any): Additional arguments to `Pip.run`

        Returns: Nothing, called for its side effect only.

        """
        self.run(cmd, check=True, **kwargs)

    def run(
        self,
        cmd: list[str],
        *,
        env: dict[str, str] | None = None,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        """Run a command within the pip virtual environment.

        Args:
            cmd: The command to run

            env: Environment variables, passed into `subprocess_run`.
                We will add additional environment variables to
                control the virtual environment activation.  Note that
                `PATH` cannot be safely set through `env` yet,
                because we have to modify that to activate the virtual
                environment, and because `subprocess.Popen` requires
                the `PATH` to be set **before** finding the program to
                call on Windows.  We may improve this in future.

            **kwargs (Any): Keyword arguments to `subprocess_run`.

        Returns: Details about the process, if `check=True` is not
           present in `kwargs`
        """
        # If the user sets a PATH, within 'env' then we will clobber
        # that when we add our envvars to the dictionary.  Later we
        # can inspect 'env' for PATH and join them together, but it's
        # not obvious what the priority should really be.
        #
        # There's another subtlety about setting PATH; see the See the
        # Warning in
        # https://docs.python.org/3/library/subprocess.html#popen-constructor
        #
        # > For Windows, ... env cannot override the PATH environment
        # > variable. Using a full path avoids all of these
        # > variations.
        #
        # The other way of doing this would be shutil.which and
        # updating the command, but that feels worse because it
        # requires that the first line of the cmd is definitely the
        # program under executation (probably reasonable) and it will
        # require logic around only doing that if a relative path is
        # given, etc.
        env = (env or {}) | self._envvars()
        with transient_envvars({"PATH": env["PATH"]}):
            return subprocess_run(cmd, env=env, **kwargs)

    def _envvars(self) -> dict[str, str]:
        base = self.path()
        path_env = base / self._venv_bin_dir()
        path = f"{path_env}{os.pathsep}{os.environ['PATH']}"
        return {"VIRTUAL_ENV": str(base), "PATH": path}

    def _auto(self) -> list[str]:
        if Path("pyproject.toml").exists():
            return ["pip", "install", "--verbose", "."]
        if Path("requirements.txt").exists():
            return ["pip", "install", "--verbose", "-r", "requirements.txt"]
        msg = "Can't determine install command"
        raise Exception(msg)

    def _venv_bin_dir(self) -> str:
        return "Scripts" if self.platform.system == "windows" else "bin"
